"""API extract, explore and export.

"""

import sys, inspect, os.path
from epydoc.markup.epytext import parse
from epydoc.markup import ParseError
from snakes.data import iterate
from snakes.hashables import hset

log = sys.stdout

def _test (x, y=None, *l, **options) :
    """A function to perform various tests.

    >>> 1+1
    2
    >>> 2+2
    4

    @param x: describe parameter C{x}
    @type x: C{xtype} or C{None}
    @param y: describe parameter C{y}
    @type y: C{ytype}
    @param l: describe parameter C{l}
    @type l: C{ltype}
    @keyword foo: describe keyword parameter C{foo}
    @type foo: C{footype}
    @keyword bar: describe keyword parameter C{bar}
    @type bar: C{bartype}
    """
    pass

class Comment (dict) :
    """
    >>> a = Comment('''#APIX# hello
    ... #APIX# foo: 1
    ... #APIX# bar: 1, 2, 3
    ... #APIX# egg x: int
    ... #APIX# egg y: int, float''')
    >>> a
    {'egg': {'y': 'int, float', 'x': 'int'},
     'foo': '1', 'bar': '1, 2, 3', 'hello': True}
    """
    def __init__ (self, comments) :
        for line in (l.strip()[6:].strip()
                     for l in comments.split("\n")
                     if l.strip().startswith("#APIX#")) :
            if ":" not in line :
                self[line] = True
            else :
                key, value = (x.strip() for x in line.split(":", 1))
                if " " in key :
                    tag, name = (x.strip() for x in key.split(" ", 1))
                    if not tag in self :
                        self[tag] = {}
                    if hasattr(self, "_read_" + tag) :
                        self[tag][name] = getattr(self, "_read_" + tag)(value.strip())
                    else :
                        self[tag][name] = value.strip()
                else :
                    if hasattr(self, "_read_" + key) :
                        self[key] = getattr(self, "_read_" + key)(value.strip())
                    else :
                        self[key] = value.strip()
    def _read_map (self, value) :
        """
        >>> Comment('#APIX# map p: x:int, y:float -> (x, y)')
        """
        proto, call = (x.strip()
                       for x in value.split("->", 1)
                       if x.strip() != "")
        return (tuple(tuple(x.strip().split(":"))
                      for x in proto.split(",")
                      if x.strip() != ""),
                call)
    def update (self, other) :
        """
        >>> a = Comment('''#APIX# hello
        ... #APIX# foo: 1
        ... #APIX# bar: 1, 2, 3
        ... #APIX# egg x: int
        ... #APIX# egg y: int, float''')
        >>> b = Comment('''#APIX# world
        ... #APIX# bar: 2, 3, 4
        ... #APIX# spam: 5
        ... #APIX# egg y: str
        ... #APIX# egg z: int''')
        >>> a.update(b)
        >>> a
        {'bar': '2, 3, 4', 'spam': '5', 'world': True, 'foo': '1',
         'egg': {'y': 'str', 'x': 'int', 'z': 'int'}, 'hello': True}
        """
        for key, val in other.items() :
            if type(val) is not dict:
                self[key] = val
            elif key not in self :
                self[key] = val.copy()
            else :
                self[key].update(val)

class Epydoc (object) :
    def __init__ (self, doc) :
        """
        >>> e = Epydoc(inspect.getdoc(_test))
        >>> e.doctest is not None
        True
        >>> e.keywords
        set(['foo', 'bar'])
        >>> e.type
        {'y': set(['ytype']), 'x': set(['None', 'xtype']),
         'foo': set(['footype']), 'bar': set(['bartype']),
         'l': set(['ltype'])}
        """
        self.doctest = None
        self.type = {}
        self.keywords = set()
        try :
            self._visit(parse(doc))
        except ParseError :
            pass
    def update (self, other) :
        if self.doctest is None :
            self.doctest = other.doctest
        for name, values in other.type.items() :
            if name in self.type :
                self.type[name].update(values)
            else :
                self.type[name] = set(values)
        self.keywords.update(other.keywords)
    def _node_doctestblock (self, tree) :
        self.doctest = tree.children[0]
    def _node_field (self, tree) :
        if tree.children[0].children[0] == "keyword" :
            self.keywords.add(tree.children[1].children[0])
            return
        elif tree.children[0].children[0] == "type" :
            name = tree.children[1].children[0]
            idx = 2
        elif tree.children[0].children[0] == "rtype" :
            name = "return"
            idx = 1
        else :
            return
        for child in tree.children[idx].children :
            try :
                if child.tag == "code" :
                    if name not in self.type :
                        self.type[name] = set()
                    self.type[name].update([child.children[0]])
            except AttributeError :
                pass
    def _visit (self, tree) :
        if isinstance(tree, str) :
            pass
        elif hasattr(self, "_node_" + tree.tag) :
            getattr(self, "_node_" + tree.tag)(tree)
        else :
            for child in tree.children :
                self._visit(child)

class ApiElement (object) :
    _instances = {}
    api = {}
    def __init__ (self, obj) :
        """
        >>> api = ApiElement(ApiElement)
        >>> api.api
        {'snakes.apix.ApiElement': <ApiElement snakes.apix.ApiElement>}
        """
        self._instances[obj] = self
        self.kind = self.__class__.__name__
        self._obj = obj
        self.api[self.full_name] = self
    def __getattr__ (self, name) :
        for cls in inspect.getmro(self.__class__) :
            if name in cls.__dict__ :
                return cls.__dict__[name]
        attr = "_api_" + name
        for cls in inspect.getmro(self.__class__) :
            if attr in cls.__dict__ :
                result = cls.__dict__[attr](self)
                self.__dict__[name] = result
                return result
        raise AttributeError, "type object '%s' has no attribute '%s'" \
              % (self.__class__.__name__, name)
    def __repr__ (self) :
        return "<%s %s>" % (self.kind, self.full_name)
    def _api_source_file (self) :
        """
        >>> ApiElement(ApiElement).source_file
        'apix.py'
        """
        try :
            srcfile = inspect.getsourcefile(self._obj)
        except TypeError :
            return None
        if srcfile is None :
            return None
        path = []
        while srcfile != "" :
            srcfile, tail = os.path.split(srcfile)
            if tail == "snakes" :
                break
            path.append(tail)
        return os.path.join(*reversed(path))
    def _api_epydoc (self) :
        """Extract information from Epydoc docstring.

        The keyword C{dummy} is not really accepted but has been
        inserted into the docstring for test purpose.

        >>> e = ApiElement(ApiElement._api_epydoc).epydoc
        >>> e.doctest is not None
        True
        >>> e.keywords
        set(['dummy'])
        >>> e.type
        {'dummy': set(['hello world!'])}

        @keyword dummy: just added for test purpose
        @type dummy: C{hello world!}
        """
        doc = self._api_doc()
        if doc is None :
            doc = ""
        return Epydoc(doc)
    def _api_name (self) :
        """
        >>> ApiElement(ApiElement).name
        'ApiElement'
        >>> ApiElement(ApiElement._api_name).name
        '_api_name'
        """
        return self._obj.__name__
    def _api_short_name (self) :
        """
        >>> ApiElement(ApiElement).short_name
        'ApiElement'
        >>> ApiElement(ApiElement._api_short_name).short_name
        'ApiElement._api_short_name'
        """
        if inspect.ismethod(self._obj) :
            return ".".join((self._obj.im_class.__name__,
                             self._api_name()))
        else :
            return self._api_name()
    def _api_source_line (self) :
        """
        >>> ApiElement(ApiElement).source_line > 0
        True
        """
        try :
            return inspect.getsourcelines(self._obj)[1]
        except TypeError :
            return None
    def _api_source_code (self) :
        """
        >>> ApiElement(ApiElement).source_code.startswith('class ApiElement (object) :')
        True
        """
        try :
            return "".join(inspect.getsourcelines(self._obj)[0])
        except TypeError :
            return None
    def _api_module (self) :
        """
        >>> ApiElement(ApiElement).module
        'snakes.apix'
        """
        try :
            return self._obj.__module__
        except AttributeError :
            return None
    def _api_doc (self) :
        """hello

        >>> ApiElement(ApiElement._api_doc).doc.startswith('hello')
        True
        """
        return inspect.getdoc(self._obj)
    def _api_full_name (self) :
        """
        >>> ApiElement(ApiElement).full_name
        'snakes.apix.ApiElement'
        >>> ApiElement(ApiElement._api_full_name).full_name
        'snakes.apix.ApiElement._api_full_name'
        """
        if inspect.ismethod(self._obj) :
            return ".".join((self._api_module(),
                             self._obj.im_class.__name__,
                             self._api_name()))
        else :
            module = self._api_module()
            if module is None :
                return self._api_name()
            else :
                return ".".join((module, self._api_name()))
    def log (self, msg) :
        print >>log, "! %s (%s:%s) - %s" % (self.full_name,
                                            self.source_file,
                                            self.source_line,
                                            msg.strip())
    def _api_comments (self) :
        """
        >>> #APIX# foo
        ... #APIX# bar: 5
        ... #APIX# egg: x=5, y=6
        ... def f () :
        ...     pass
        >>> ApiElement(f).comments
        {'egg': 'x=5, y=6', 'foo': True, 'bar': '5'}
        """
        com = inspect.getcomments(self._obj)
        if com is None :
            return {}
        else :
            return Comment(com)

class Function (ApiElement) :
    _instances = {}
    def _api_arguments (self) :
        """
        >>> Function(_test).arguments
        ['x', 'y']
        """
        try :
            return inspect.getargspec(self._obj)[0]
        except TypeError :
            return None
    def _api_list_argument (self) :
        """
        >>> Function(_test).list_argument
        'l'
        """
        try :
            return inspect.getargspec(self._obj)[1]
        except TypeError :
            return None
    def _api_keyword_argument (self) :
        """
        >>> Function(_test).keyword_argument
        'options'
        """
        try :
            return inspect.getargspec(self._obj)[2]
        except TypeError :
            return None
    def _api_default (self) :
        """
        >>> Function(_test).default
        {'y': None}
        """
        try :
            args, larg, karg, default = inspect.getargspec(self._obj)
        except TypeError :
            return None
        if default is None :
            return None
        else :
            return dict(zip(reversed(args), reversed(default)))
    def _api_type (self) :
        result = {}
        result.update(self.epydoc.type)
        for key, val in self.comments.get("type", {}).items() :
            result[key] = set(t.strip()
                              for t in val.split(",")
                              if t.strip() != "")
        return result
    def _api_keywords (self) :
        result = set()
        result.update(self.epydoc.keywords)
        result.update(k.strip()
                      for k in self.comments.get("keywords", "").split(",")
                      if k.strip() != "")
        return result
    def _api_signature (self) :
        result = ", ".join("%s:%s" % (a, "|".join(self.type[a]))
                           for a in self.arguments)
        if self.list_argument is not None :
            result += ", *%s:[%s]" % (self.list_argument,
                                      "|".join(self.type[self.list_argument]))
        if self.keyword_argument is not None :
            if len(self.keywords) > 0 :
                result += ", **%s:{%s}" % (self.keyword_argument,
                                           ",".join("%s:%s" % (k, "|".join(self.type[k]))
                                                    for k in self.keywords))
            else :
                result += ", **%s:{str:object}" % self.keyword_argument
        return "(%s):%s" % (result,
                            "|".join(self.type.get("return", ["None"])))
    def _api_exported_arguments (self) :
        karg = self.keyword_argument is not None
        if karg and len(self.keywords) == 0 :
            return self.arguments, self.list_argument, self.keyword_argument
        amap = self.comments.get("map", {})
        old = list(self.arguments)
        if karg :
            old.extend(self.keywords)
        new = []
        for arg in old :
            if arg in amap :
                for n, t in amap[arg][0] :
                    self.type[n] = set([t])
                    new.append(n)
            else :
                new.append(arg)
        return tuple(new), self.list_argument, None
    def _api_exported_call (self) :
        karg = self.keyword_argument is not None
        if karg and len(self.keywords) == 0 :
            return self.arguments, self.list_argument, self.keyword_argument
        amap = self.comments.get("map", {})
        pargs = []
        for arg in self.arguments :
            if arg in amap :
                pargs.append(amap[arg][1])
            else :
                pargs.append(arg)
        if karg :
            kargs = {}
            for arg in self.keywords :
                if arg in amap :
                    kargs[arg] = amap[arg][1]
                else :
                    kargs[arg] = arg
            return tuple(pargs), self.list_argument, kargs
        else :
            return tuple(pargs), self.list_argument, None


class _TestA (object) :
    #APIX# hello
    #APIX# foo: 1
    #APIX# bar: 1, 2, 3
    #APIX# egg x: int
    #APIX# egg y: int, float
    def method (self, x, y) :
        """
        @type x: C{int}
        @type y: C{int}
        """
    def foo (self) :
        pass

class _TestB(_TestA) :
    #APIX# world
    #APIX# foo: 2
    #APIX# bar: 2, 3, 4
    #APIX# spam: 5
    #APIX# egg x: str, object
    #APIX# egg y: str
    #APIX# egg z: int
    def method (self, x, y, **options) :
        """
        @keyword foo: ...
        @type foo: C{int}
        @keyword bar: ...
        @type bar: C{int}
        """
    def bar (self) :
        pass
    x = 5

class Method (Function) :
    def _api_epydoc (self) :
        """
        >>> e = Method(_TestB.method).epydoc
        >>> e.type
        {'y': set(['int']), 'x': set(['int']),
         'self': 'snakes.apix._TestB',
         'foo': set(['int']), 'bar': set(['int'])}
        >>> e.keywords
        set(['foo', 'bar'])
        >>> e.doctest is None
        True
        """
        result = Function._api_epydoc(self)
        for cls in inspect.getmro(self._obj.im_class) :
            if cls is not self._obj.im_class :
                if hasattr(cls, self.name) :
                    other = Function(getattr(cls, self.name)).epydoc
                    if other is not None :
                        result.update(other)
        result.type['self'] = ApiElement(self._obj.im_class).full_name
        return result
    def _api_comments (self) :
        """
        >>> Method(_TestB.method).comments
        {'bar': '1, 2, 3', 'spam': '5', 'world': True, 'foo': '1',
         'egg': {'y': 'int, float', 'x': 'int', 'z': 'int'}, 'hello': True}
        """
        result = Function._api_comments(self)
        if result is None :
            result = Comment("")
        for cls in inspect.getmro(self._obj.im_class) :
            if cls is not self._obj.im_class :
                if hasattr(cls, self._obj.__name__) :
                    other = Function(getattr(cls, self._obj.__name__)).comments
                    if other is not None :
                        result.update(other)
        return result

class Class (ApiElement) :
    _instances = {}
    def _api_bases (self) :
        """
        >>> Class(_TestB).bases
        ['_TestA', 'object']
        """
        return [cls.__name__ for cls in inspect.getmro(self._obj)
                if cls is not self._obj]
    def _is_method (self, obj) :
        return inspect.ismethod(obj) and ((not obj.__name__.startswith("_"))
                                          or (obj.__name__.startswith("__")
                                              and obj.__name__.endswith("__")))
    def _api_methods (self) :
        """
        >>> Class(_TestB).methods
        [<Method snakes.apix._TestB.bar>,
         <Method snakes.apix._TestB.foo>,
         <Method snakes.apix._TestB.method>]
        """
        result = []
        done = set()
        for cls in inspect.getmro(self._obj) :
            for name, value in inspect.getmembers(cls, self._is_method) :
                if name not in done :
                    m = Method(value)
                    if cls not in Class._instances :
                        Class(cls)
                    m.container = Class._instances[cls]
                    result.append(m)
                    done.add(name)
        return result
    def _is_not_method (self, obj) :
        return not inspect.ismethod(obj)
    def _api_objects (self) :
        """
        >>> Class(_TestB).objects
        {'x': <Object snakes.apix._TestB.x>}
        """
        result = {}
        for cls in inspect.getmro(self._obj) :
            for name, value in inspect.getmembers(cls, self._is_not_method) :
                if not (name in result or name.startswith("_")) :
                    if cls not in Class._instances :
                        Class(cls)
                    result[name] = Object(name, value, Class._instances[cls])
        return result

class Module (ApiElement) :
    _instances = {}
    def _api_module (self) :
        return None
    def _api_name (self) :
        return self._obj.__name__
    def _api_full_name (self) :
        return self._obj.__name__
    def _api_classes (self) :
        """
        >>> import snakes.data
        >>> Module(snakes.data).classes
        [<Class snakes.DomainError>, <Class snakes.data.MultiSet>,
         <Class snakes.data.Substitution>, <Class snakes.data.WordSet>,
         <Class snakes.hashables.hdict>]
        """
        return [Class(v) for n, v in inspect.getmembers(self._obj,
                                                        inspect.isclass)
                if not n.startswith("_")]
    def _is_object (self, obj) :
        return not (inspect.isclass(obj)
                    or inspect.isfunction(obj)
                    or inspect.ismodule(obj)
                    or inspect.isbuiltin(obj))
    def _api_objects (self) :
        return [Object(n, v, self)
                for n, v in inspect.getmembers(self._obj, self._is_object)
                if not n.startswith("_")]
    def _api_functions (self) :
        """
        >>> import snakes.data
        >>> Module(snakes.data).functions
        [<Function snakes.data.cross>, <Function snakes.data.iterate>,
         <Function snakes.pnml.obj_to_pnml>,
         <Function snakes.pnml.pnml_to_obj>,
         <Function snakes.pnml.pnmldump>,
         <Function snakes.pnml.pnmlload>,
         <Function snakes.pnml.string>]
        """
        return [Function(v) for n, v in inspect.getmembers(self._obj,
                                                           inspect.isfunction)
                if not n.startswith("_")]

class Object (ApiElement) :
    _instances = {}
    def __init__ (self, name, obj, container) :
        self.name = name
        self._container = container
        ApiElement.__init__(self, obj)
    def _api_type (self) :
        """
        >>> Object('x', _TestB.x, Class(_TestB)).type
        'int'
        """
        return self._obj.__class__.__name__
    def _api_name (self) :
        """
        >>> Object('x', _TestB.x, Class(_TestB)).name
        'x'
        """
        return self.name
    def _api_full_name (self) :
        """
        >>> Object('x', _TestB.x, Class(_TestB)).full_name
        'snakes.apix._TestB.x'
        """
        return ".".join([self._container.full_name, self.name])
    def _api_module (self) :
        """
        >>> Object('x', _TestB.x, Class(_TestB)).module
        'snakes.apix'
        """
        return self._container.module
    def _api_container (self) :
        """
        >>> Object('x', _TestB.x, Class(_TestB)).container
        'snakes.apix._TestB'
        """
        return self._container.full_name

class Export (object) :
    def __init__ (self, api, name) :
        self.name = name
        self.api = api
        self._name = {}
        self._type_map = self.__class__._type_map.copy()
    def log (self, msg) :
        print >>log, "! %s: %s" % (self.__class__.__name__, msg.strip())
    def map_name (self, api) :
        """
        >>> api = Class(Export)
        >>> exp = Export(api, 'snk')
        >>> exp.map_name(Method(Export.__init__))
        'snk_new_Export'
        >>> exp.map_name(Method(Export.map_name))
        'snk_Export_map_name'
        """
        if api.full_name not in self._name :
            parts = [self.name] + api.full_name[len(api.module)+1:].split(".")
            if parts[-1] == "__init__" :
                parts[-2:] = ["new", parts[-2]]
            self._name[api.full_name] = "_".join(parts)
        return self._name[api.full_name]
    def best_type (self, *typesets) :
        """
        >>> api = Method(Export.best_type)
        >>> exp = Export(api, 'snk')
        >>> exp.best_type(['int', 'float'])
        'float'
        >>> exp.best_type(['float', 'int', 'bool'])
        'float'
        >>> exp.best_type(['float', 'int', 'bool', 'list'])
        'object'
        >>> exp.best_type(['dict'])
        'dict'
        >>> exp.best_type(['None', 'str'])
        'str'
        >>> exp.best_type(['None', 'int'])
        'object'
        """
        s = set(iterate(typesets[0]))
        for t in typesets[1:] :
            s.update(iterate(t))
        for x, y in (("bool", "int"),
                     ("long", "int"),
                     ("int", "float"),
                     ("None", "str")) :
            if y in s :
                s.discard(x)
        if len(s) == 1 :
            return s.pop()
        else :
            return "object"
    _type_map = {"int" : "int",
                 "object" : "object",
                 "str" : "char*",
                 "float" : "float",
                 "bool" : "int",
                 "long" : "int",
                 "None" : "void"}
    def map_type (self, types) :
        """
        >>> api = Class(Export)
        >>> exp = Export(api, 'snk')
        >>> exp.map_type(['int'])
        'int'
        >>> exp.map_type(['str'])
        'char*'
        >>> exp.map_type(['int', 'float'])
        'float'
        >>> exp.map_type(['float', 'int', 'bool'])
        'float'
        >>> exp.log = lambda x:None # dirty trick to suppress warnings
        >>> exp.map_type(['dict'])
        'object'
        >>> exp.map_type(['None', 'str'])
        'char*'
        >>> exp.map_type(['None', 'int'])
        'object'
        """
        if len(types) == 0 :
            raise ValueError, "empty type specification"
        elif isinstance(types, str) :
            types = hset([types])
        else :
            types = hset(types)
        if types not in self._type_map :
            norm = self.best_type(types)
            if norm in self._type_map :
                self._type_map[types] = self._type_map[norm]
            else :
                self._type_map[types] = "object"
            if (self._type_map[types] == "object") and ("object" not in types) :
                self.log("casting '%s' to 'object'" % "|".join(types))
        return self._type_map[types]
    def export (self, pyrex=[], swig=[]) :
        for ext, lines in (("pyx", pyrex), ("i", swig)) :
            self._lines = lines[:]
            getattr(self, "_export_%s_%s" % (api.kind, ext))()
            out = open("%s.%s" % (self.name, ext), "w")
            out.writelines(l + "\n" for l in self._lines)
            out.close()
            hfile, hlines = getattr(self, "_export_Types_%s" % ext)()
            out = open(target, "w")
            out.writelines(l + "\n" for l in hlines)
            out.close()
    def _export_Types_pyx (self) :
        hfile = []
        parts = ["# va_list helpers", 'cdef extern from "%sargs.h":' % self.name]
        for key, val in self._type.items() :
            if isinstance(key, str) :
                parts.append("    %sva_%s(va_list ap)" % (val, key))
                if val == "object" :
                    val = "PyObject*"
                hfile.append("#define va_%s(ap) va_arg(ap, %s)" % (key, val))
        self._lines[:0] = parts
        return "%sargs.h" % self.name, hfile
    def _export_Module_pyx (self, api) :
        self._lines.append("# objects")
        for obj in api.objects :
            if skip not in obj.comments :
                self._lines.append("cdef %s %s" % (self.map_type(obj.type),
                                                   self.map_name(obj)))
                self._lines.append("%s = %s.%s" % (self.map_name(obj),
                                                   self.name,
                                                   obj.short_name))
        self._lines.append("# functions")
        for fun in api.functions :
            if "skip" not in fun.comments :
                self._lines.append(self._export_Function_pyx(fun))
        for cls in api.classes :
            if "skip" not in cls.comments :
                self._lines.append("# class %s" % cls.full_name)
                for fun in cls.methods :
                    if "skip" not in fun.comments :
                        self._lines.append("# method %s" % fun.fullname)
                        self._lines.append(self._export_Method_pyx(fun))
    def prototype (self, api) :
        args, larg, karg = api.exported_arguments
        proto = ", ".join("%s %s" % (self.map_type(api.type[a]), a)
                          for a in args)
        if larg is not None or karg is not None :
            return proto + ", ..."
        else :
            return proto
    def call (self, api) :
        args, larg, karg = api.exported_call
        result = list(args)
        if larg is not None :
            result.append("*" + larg)
        if isinstance(karg, str) :
            result.append("**" + karg)
        elif isinstance(karg, dict) :
            result.extend("%s=%s" % (k, v) for k, v in karg.items())
        return ", ".join(result)
    def _export_Function_pyx (self, api) :
        body = ["# %(full_name)s%(signature)s",
                "cdef public %(rtype)s %(map_name)s (%(proto)s) :"]
        ktype = None
        cktype = None
        vtype = None
        ltype = None
        cltype = None
        a, l, k = api.exported_arguments
        larg = l is not None
        karg = k is not None
        if larg :
            ltype = self.best_type(api.type[api.list_argument])
            cltype = self.map_type(ltype)
            body.extend(["    cdef va_list arglist",
                         "    cdef %(cltype)s _%(larg)s_item",
                         "    %(larg)s = []"])
        if karg :
            ktype = "str"
            cktype = self.map_type(ktype)
            if len(api.keywords) == 0 :
                vtype = self.map_type(["object"])
            else :
                vtype = self.map_type(self.best_type(*(api.type[k]
                                                       for k in api.keywords)))
            if api.list_argument is None :
                body.append("    cdef va_list arglist")
            body.extend(["    cdef %(cktype)s _%(karg)s_key",
                         "    cdef %(vtype)s _%(karg)s_val",
                         "    %(karg)s = {}"])
        rtype = self.map_type(api.type.get("return", "None"))
        if rtype == "char*" :
            body.extend(["    global _str_return",
                         "    cdef char *_%(map_name)s_ret"])
        body.append("    try :")
        if larg :
            body.extend(["        va_start(arglist, <void*>%(last_arg)s)",
                         "        while True :",
                         "            _%(larg)s_item = va_%(ltype)s(arglist)",
                         "            if _%(larg)s_item is NULL :",
                         "                break",
                         "            %(larg)s.append(_%(larg)s_item)"])
            if not karg :
                body.append("        va_end(arglist)")
        if karg :
            if not larg :
                body.append("        va_start(arglist, <void*>%(last_arg)s)")
            body.extend(["        while True :",
                         "            _%(karg)s_key = va_%(ktype)s(arglist)",
                         "            if _%(karg)s_key is NULL :",
                         "                break",
                         "            _%(karg)s_val = va_%(vtype)s(arglist)",
                         "            %(karg)s[_%(karg)s_key] = _%(karg)s_val",
                         "        va_end(arglist)"])
        if rtype == "char*" :
            body.extend(["        _str_return = %(module)s.%(name)s(%(args)s)",
                         "        _%(map_name)s_ret = _str_return",
                         "        return _%(map_name)s_ret"])
        elif rtype == "void" :
            body.append("        %(module)s.%(name)s(%(args)s)")
        else :
            body.append("        return %(module)s.%(name)s(%(args)s)")
        body.extend(["    except :",
                     "        _save_error()"])
        d = {"rtype" : rtype,
             "module" : self.name,
             "name" : api.name,
             "short_name" : api.short_name,
             "full_name" : api.full_name,
             "map_name" : self.map_name(api),
             "larg" : api.list_argument,
             "karg" : api.keyword_argument,
             "vtype" : vtype,
             "ltype" : ltype,
             "cltype" : cltype,
             "ktype" : ktype,
             "cktype" : cktype,
             "signature" : api.signature,
             "last_arg" : api.exported_arguments[0][-1],
             "args" : self.call(api),
             "proto" : self.prototype(api)}
        self._lines.extend(l % d for l in body)
        self._lines.append("")
    def _export_Method_pyx (self, api) :
        pass
    def _export_Class_pyx (self, api) :
        pass
    def _export_Object_pyx (self, api) :
        pass
    def _export_Module_i (self, api) :
        pass
    def _export_Function_i (self, api) :
        pass
    def _export_Method_i (self, api) :
        pass
    def _export_Class_i (self, api) :
        pass
    def _export_Object_i (self, api) :
        pass
