"""This module allows to parse Python code in order to perform various
operations on it. The main application is to rename variables, another
is to translate Python expressions into other langages.

The base class C{Parser} is useless unless it is derived:

>>> parser = Parser()
>>> print parser('f(x)+2==3')
(((f)(x)) + (2)) == (3)

A translator to the syntax of C/C++ is provided:

>>> trans = TranslateToC()
>>> print trans('(x==1 and y==2) or not (x+1==y+2)')
((x) == (1)) && ((y) == (2)) || ! (((x) + (1)) == ((y) + (2)))

Renaming variables is easy and will not replace variable names within
strings:

>>> renamer = Renamer({'x': 'y', 'y': 'z'})
>>> print renamer('x+1 == y+2 and len("x, y, z are strings!") == 20')
(((y) + (1)) == ((z) + (2))) and (((len)('x, y, z are strings!')) == (20))

Renaming and translation can be combined:

>>> renamer = Renamer({'x': 'y', 'y': 'z'}, parser=TranslateToC)
>>> print renamer("x+1 == y+2 and len('x, y, z are strings!') == 20")
(((y) + (1)) == ((z) + (2))) && ((len("x, y, z are strings!")) == (20))

Finding variables involved in an expression is possible also, the
names of the called function are not considered as variables:

>>> finder = Variables()
>>> list(sorted(finder("x+1 == y+2 and len('x, y, z are strings!') == 20")))
['x', 'y']
"""

import sys, compiler

class ParserError (Exception) :
    """The exception raised when an expression cannot be parsed or
    translated.

    >>> parser = Parser()
    >>> print parser("x+4 = 3") # invalid Python syntax
    Traceback (most recent call last):
      ...
    ParserError: syntax error (can't assign to operator)
    >>> print parser("def f(x) : pass") # unsupported construct
    Traceback (most recent call last):
      ...
    ParserError: function definition not supported
    >>> print parser("x+1 == y+2 == 20") # no error in Python
    ((x) + (1)) == ((y) + (2)) == (20)
    >>> parser = TranslateToC()
    >>> print parser("x+1 == y+2 == 20") # not supported in C
    Traceback (most recent call last):
      ...
    ParserError: chained comparison not supported in C
    """

class Parser (object) :
    "The base class for parsing Python expressions."
    _unsupported = {"AssAttr" : "assignment",
                    "AssList" : "assignment",
                    "Assert" : "assertion",
                    "Assign" : "assignment",
                    "AugAssign" : "assignment",
                    "Break" : "break",
                    "Class" : "class definition",
                    "Continue" : "continue",
                    "Decorators" : "decorators",
                    "Ellipsis" : "ellipsis",
                    "Exec" : "exec",
                    "For" : "loop",
                    "From" : "import",
                    "Function" : "function definition",
                    "Global" : "global statement",
                    "If" : "if statement",
                    "Import" : "import",
                    "Lambda" : "lambda functions",
                    "Pass" : "pass statement",
                    "Print" : "print statement",
                    "Printnl" : "print statement",
                    "Raise" : "raise statement",
                    "Return" : "return statement",
                    "TryExcept" : "try statement",
                    "TryFinally" : "try statement",
                    "While" : "loop",
                    "Yield" : "yield statement"}
    def __call__ (self, expr) :
        """Parse a Python statement.

        @param expr: the expression or statement to parse
        @type expr: C{str}
        @return: the result of the parsing in canonical form
        @rtype: C{str}
        """
        try :
            ast = compiler.parse(expr)
        except :
            raise ParserError, "syntax error (%s)" % str(sys.exc_info()[1])
        return self[ast]
    #C# skip
    def __getitem__ (self, node) :
        name = self._name_(node)
        return getattr(self, "_%s_" % name)(node)
    def _name_ (self, node) :
        name = node.__class__.__name__
        if name in self._unsupported :
            raise ParserError, "%s not supported" % self._unsupported[name]
        elif not hasattr(self, "_%s_" % name) :
            raise ParserError, "unsupported construct '%s'" % name
        return name
    def _Add_ (self, node) :
        "addition"
        return "(%s) + (%s)" % (self[node.left], self[node.right])
    def _And_ (self, node) :
        "boolean and"
        return " and ".join(["(%s)" % self[n] for n in node.nodes])
    def _AssName_ (self, node) :
        "assignment"
        return node.name
    def _AssTuple_ (self, node) :
        "tuple assignment"
        return ", ".join([self[n] for n in node.nodes])
    def _Backquote_ (self, node) :
        "back-quotes"
        return "repr(%s)" % self[node.expr]
    def _Bitand_ (self, node) :
        "binary and"
        return " & ".join(["(%s)" % self[n] for n in node.nodes])
    def _Bitor_ (self, node) :
        "binary or"
        return " | ".join(["(%s)" % self[n] for n in node.nodes])
    def _Bitxor_ (self, node) :
        "binary xor"
        return " ^ ".join(["(%s)" % self[n] for n in node.nodes])
    def _CallFunc_ (self, node) :
        "function call"
        args = [self[n] for n in node.args]
        if node.star_args is not None :
            args.append(self[node.star_args])
        if node.dstar_args is not None :
            args.append(self[node.dstar_args])
        return "(%s)(%s)" % (self[node.node], ", ".join(args))
    def _Compare_ (self, node) :
        "comparison"
        parts = []
        for e, n in (((i % 2) == 0, n)
                           for i, n in enumerate(node.getChildren())) :
            if e :
                parts.append("(%s)" % self[n])
            else :
                parts.append(n)
        return " ".join(parts)
    def _Const_ (self, node) :
        "constant"
        return repr(node.value)
    def _Dict_ (self, node) :
        "dictionnary"
        return repr(dict(((self[key], self[val]) for key, val in node.items)))
    def _Discard_ (self, node) :
        "function call (discarding result)"
        return self[node.expr]
    def _Div_ (self, node) :
        "division"
        return "(%s) / (%s)" % (self[node.left], self[node.rigth])
    def _Getattr_ (self, node) :
        "attribute lookup"
        return "(%s).%s" % (self[node.expr], node.attrname)
    def _Invert_ (self, node) :
        "binary negation"
        return "~(%s)" % self[node.expr]
    def _Keyword_ (self, node) :
        "keyword parameter"
 	return "%s=(%s)" % (node.name, self[node.expr])
    def _LeftShift_ (self, node) :
        "binary left shift"
 	return "(%s) << (%s)" % (self[node.left], self[node.right])
    def _List_ (self, node) :
        "list"
        return str(["(%s)" % self[n] for n in node.nodes])
    def _ListComp_ (self, node) :
        "list comprehension"
        return "[%s %s]" % (self[node.expr],
                            " ".join([self[q] for q in node.quals]))
    def _ListCompFor_ (self, node) :
        "loop (in list comprehension)"
        if len(node.ifs) == 0 :
            return "for %s in (%s)" % (self[node.assign], self[node.list])
        else :
            return "for %s in (%s) %s" % (self[node.assign],
                                          self[node.list],
                                          "".join([self[i] for i in node.ifs]))
    def _ListCompIf_ (self, node) :
        "test (in list comprehension)"
        return "if (%s)" % self[node.test]
    def _Mod_ (self, node) :
        "modulo"
        return "(%s) %% (%s)" % (self[node.left], self[node.right])
    def _Module_ (self, node) :
        "module content"
        return self[node.node]
    def _Mul_ (self, node) :
        "multiplication"
        return "(%s) * (%s)" % (self[node.left], self[node.right])
    def _Name_ (self, node) :
        "variable name"
        return node.name
    def _Not_ (self, node) :
        "boolean negation"
        return "not (%s)" % self[node.expr]
    def _Or_ (self, node) :
        "boolean or"
        return " or ".join([self[n] for n in node.nodes])
    def _Power_ (self, node) :
        "exponentiation"
        return "(%s) ** (%s)" % (self[node.left], self[node.right])
    def _RightShift_ (self, node) :
        "binary right shift"
 	return "(%s) >> (%s)" % (self[node.left], self[node.right])
    def _Slice_ (self, node) :
        "simple list slice"
        if (node.lower is None) and (node.upper is None) :
            return "(%s)[:]" % self[node.expr]
        elif node.upper is None :
            return "(%s)[%s:]" % (self[node.expr], self[node.lower])
        elif node.lower is None :
            return "(%s)[:%s]" % (self[node.expr], self[node.upper])
        else :
            return "(%s)[%s:%s]" % (self[node.expr],
                                    self[node.lower],
                                    self[node.upper])
    def _Sliceobj_ (self, node) :
        "extended list slice"
        parts = []
        for x in (self[n] for n in node.nodes) :
            if x == "None" :
                parts.append("")
            else :
                parts.append(x)
        return ":".join(parts)
    def _Stmt_ (self, node) :
        "statement"
        if len(node.nodes) > 1 :
            raise ParserError, "multiple statements forbidden"
        return self[node.nodes[0]]
    def _Sub_ (self, node) :
        "substraction"
        return "(%s) - (%s)" % (self[node.left], self[node.right])
    def _Subscript_ (self, node) :
        "container lookup"
        if len(node.subs) > 1 :
            return "(%s)[%s]" % (self[self.expr],
                                 ", ".join([self[s] for s in node.subs]))
        else :
            return "(%s)[%s]" % (self[self.expr], self[node.subs[0]])
    def _Tuple_ (self, node) :
        "tuple"
        return str(tuple("(%s)" % self[n] for n in node.nodes))
    def _UnaryAdd_ (self, node) :
        "arithmetic positive"
        return "+(%s)" % self[node.expr]
    def _UnarySub_ (self, node) :
        "arithmetic negation"
        return "-(%s)" % self[node.expr]

#C# ignore: parser
def Renamer (convert, parser=Parser) :
    """Construct a C{Parser} that allows for renaming variables.

    @param convert: the dictionnary for renamings
    @type convert: C{{str: str}}
    """
    class Renamer (parser) :
        def _Name_ (self, node) :
            return convert.get(node.name, node.name)
        def _AssName_ (self, node) :
            return convert.get(node.name, node.name)
    return Renamer()

#C# ignore: parser
def Variables (parser=Parser) :
    "Construct a C{Parser} that searches for variables."
    class Variables (parser) :
        def __init__ (self) :
            self._vars = set()
        def _Name_ (self, node) :
            if node.name not in ('True', 'False', 'None') :
                self._vars.add(node.name)
            return node.name
        def _AssName_ (self, node) :
            self._vars.add(node.name)
            return node.name
        def _CallFunc_ (self, node) :
            args = [self[n] for n in node.args]
            if node.star_args is not None :
                args.append(self[node.star_args])
            if node.dstar_args is not None :
                args.append(self[node.dstar_args])
            result = "(%s)(%s)" % (self[node.node], ", ".join(args))
            self._vars.discard(node.node.name)
            return result
        def __call__ (self, expr) :
            parser.__call__(self, expr)
            return self._vars
    return Variables()

class TranslateToC (Parser) :
    "Translate (some) Python expression to C syntax."
    def __init__ (self) :
        "Creates the parser"
        self._unsupported = Parser._unsupported.copy()
        for name in ["AssTuple", "Backquote", "Bitand", "Bitor", "Bitxor",
                     "Dict", "Invert", "LeftShift", "List", "ListComp",
                     "ListCompFor", "ListCompIf", "RightShift", "Slice",
                     "Sliceobj", "Keyword", "Tuple"] :
            self._unsupported[name] = getattr(Parser, "_%s_" % name).__doc__
    def _And_ (self, node) :
        "boolean and"
        return " && ".join(["(%s)" % self[n] for n in node.nodes])
    def _CallFunc_ (self, node) :
        "function call"
        args = [self[n] for n in node.args]
        if node.star_args is not None :
            raise ParserError, "*args not supported in C"
        if node.dstar_args is not None :
            raise ParserError, "**args not supported in C"
        if not isinstance(node.node, compiler.ast.Name) :
            raise ParserError, "invalid C function name '%s'" % self[node.node]
        return "%s(%s)" % (self[node.node], ", ".join(args))
    _comp = {"<>" : "!=",
             "is" : "==",
             "is not" : "!=",
             "<=" : "<=",
             ">=" : ">=",
             "<" : "<",
             ">" : ">",
             "==" : "==",
             "!=" : "!="}
    def _Compare_ (self, node) :
        "comparison"
        parts = []
        for e, n in (((i % 2) == 0, n)
                           for i, n in enumerate(node.getChildren())) :
            if e :
                parts.append("(%s)" % self[n])
            elif n in self._comp :
                parts.append(self._comp[n])
            else :
                raise ParserError, "comparison '%s' not supported in C" % n
            if len(parts) > 3 :
                raise ParserError, "chained comparison not supported in C"
        return " ".join(parts)
    def _escape (self, string) :
        def _escape_ () :
            _prev_ = None
            for char in repr(string)[1:-1] :
                if char == "'" :
                    if _prev_ != "\\" :
                        yield "\\"
                    yield "'"
                elif char == '"' :
                    if _prev_ != "\\" :
                        yield "\\"
                    yield '"'
                else :
                    yield char
                _prev_ = char
        return "".join([c for c in _escape_()])
    def _Const_ (self, node) :
        "constant"
        if isinstance(node.value, str) :
            return '"%s"' % self._escape(node.value)
        else :
            return repr(node.value)
    def _Getattr_ (self, node) :
        "struct field"
        if not isinstance(self.node.expr, compiler.Name) :
            raise ParserError, "invalid C struct name '%s'" % self[node.expr]
        return "%s.%s" % (self[node.expr], node.attrname)
    def _Not_ (self, node) :
        "boolean negation"
        return "! (%s)" % self[node.expr]
    def _Or_ (self, node) :
        "boolean or"
        return " || ".join([self[n] for n in node.nodes])
    def _Subscript_ (self, node) :
        "array lookup"
        if not isinstance(node.expr, compiler.ast.Name) :
            raise ParserError, "invalid C array name '%s'" % self[node.expr]
        elif len(node.subs) > 1 :
            return "%s(%s)" % (self[node.expr],
                               ", ".join([self[s] for s in node.subs]))
        else :
            return "%s(%s)" % (self[node.expr], self[node.subs[0]])
