"""Basic data types and functions used in SNAKES"""

import operator
from snakes import DomainError
from snakes.hashables import hdict
from snakes.pnml import pnmldump, pnmlload, obj_to_pnml, pnml_to_obj, string

def cross (sets) :
    """Cross-product.

    >>> list(cross([[1, 2], [3, 4, 5]]))
    [(1, 3), (1, 4), (1, 5), (2, 3), (2, 4), (2, 5)]
    >>> list(cross([[1, 2], [3, 4, 5], [6, 7, 8, 9]]))
    [(1, 3, 6), (1, 3, 7), (1, 3, 8), (1, 3, 9), (1, 4, 6), (1, 4, 7),
     (1, 4, 8), (1, 4, 9), (1, 5, 6), (1, 5, 7), (1, 5, 8), (1, 5, 9),
     (2, 3, 6), (2, 3, 7), (2, 3, 8), (2, 3, 9), (2, 4, 6), (2, 4, 7),
     (2, 4, 8), (2, 4, 9), (2, 5, 6), (2, 5, 7), (2, 5, 8), (2, 5, 9)]
    >>> list(cross([[], [1]]))
    []

    @param sets: the sets of values to use
    @type sets: C{iterable(iterable(object))}
    @return: the C{list} of obtained tuples (lists are used to allow
      unhashable objects)
    @rtype: C{generator(tuple(object))}
    """
    if len(sets) == 0 :
        pass
    elif len(sets) == 1 :
        for item in sets[0] :
            yield (item,)
    else :
        for item in sets[0] :
            for others in cross(sets[1:]) :
                yield (item,) + others

def iterate (value) :
    """Like Python's builtin C{iter} but consider strings as atomic.

    >>> list(iter([1, 2, 3]))
    [1, 2, 3]
    >>> list(iterate([1, 2, 3]))
    [1, 2, 3]
    >>> list(iter('foo'))
    ['f', 'o', 'o']
    >>> list(iterate('foo'))
    ['foo']

    @param value: any object
    @type value: C{object}
    @return: an iterator on the elements of C{value} if is is iterable
        and is not string, an iterator on the sole C{value} otherwise
    @rtype: C{generator}
    """
    if isinstance(value, str) :
        return iter([value])
    else :
        try :
            return iter(value)
        except TypeError :
            return iter([value])

class WordSet (set) :
    """A set of words being able to generate fresh words."""
    def fresh (self, add=False, min=1, allowed="abcdefghijklmnopqrstuvwxyz") :
        """Create a fresh word (ie, which is not in the set).

        >>> w = WordSet(['foo', 'bar'])
        >>> list(sorted(w))
        ['bar', 'foo']
        >>> w.fresh(True, 3)
        'aaa'
        >>> list(sorted(w))
        ['aaa', 'bar', 'foo']
        >>> w.fresh(True, 3)
        'baa'
        >>> list(sorted(w))
        ['aaa', 'baa', 'bar', 'foo']

        @param add: add the created word to the set if C{add=True}
        @type add: C{bool}
        @param min: minimal length of the new word
        @type min: C{int}
        @param allowed: characters allowed in the new word
        @type allowed: C{str}
        """
        result = [allowed[0]] * min
        pos = 0
        while "".join(result) in self :
            for c in allowed :
                try :
                    result[pos] = c
                except IndexError :
                    result.append(c)
                if "".join(result) not in self :
                    break
            pos += 1
        if add :
            self.add("".join(result))
        return "".join(result)

class MultiSet (hdict) :
    """Set with repetitions, ie, function from values to integers.

    MultiSets support various operations, in particular: addition
    (C{+}), substraction (C{-}), multiplication by a non negative
    integer (C{*k}), comparisons (C{<}, C{>}, etc.), length (C{len})"""
    def __init__ (self, values=[]) :
        """Initialise the multiset, adding values to it.

        @param values: a single value or an iterable object holding
          values (strings are not iterated)
        @type values: any atomic object (C{str} included) or an
          iterable object
        """
        self.add(values)
    def copy (self) :
        result = MultiSet()
        result.update(self)
        return result
    @pnmldump
    def __pnmldump__ (self, doc) :
        result = doc.createElement("multiset")
        for value in hdict.__iter__(self) :
            item = doc.createElement("item")
            item.setAttribute("multiplicity", str(self[value]))
            item.appendChild(obj_to_pnml(doc, value))
            result.appendChild(item)
        return result
    @pnmlload
    def __pnmlload__ (cls, node) :
        result = cls()
        for item in node.childNodes :
            times = int(item.getAttribute("multiplicity"))
            value = pnml_to_obj(item.firstChild)
            result.add(value, times)
        return result
    def _add (self, value, times=1) :
        """Add a single value C{times} times.

        @param value: the value to add
        @type value: any object
        @param times: the number of times that C{value} has to be
          added
        @type times: non negative C{int}
        """
        if times < 0 :
            raise ValueError, "negative values are forbidden"
        try :
            self[value] += times
        except KeyError :
            self[value] = times
    def add (self, values, times=1) :
        """Add values to the multiset.

        >>> m = MultiSet()
        >>> m.add([1, 2, 2, 3], 2)
        >>> list(sorted(m.items()))
        [1, 1, 2, 2, 2, 2, 3, 3]
        >>> m.add(5, 3)
        >>> list(sorted(m.items()))
        [1, 1, 2, 2, 2, 2, 3, 3, 5, 5, 5]

        @param values: the values to add or a single value to add
        @type values: any atomic object (C{str} included) or an
          iterable object
        @param times: the number of times each value should be added
        @type times: non negative C{int}"""
        for value in iterate(values) :
            self._add(value, times)
    def _remove (self, value, times=1) :
        """Remove a single value C{times} times.

        @param value: the value to remove
        @type value: any object
        @param times: the number of times that C{value} has to be
          removed
        @type times: non negative C{int}
        """
        if times < 0 :
            raise ValueError, "negative values are forbidden"
        if times > self.get(value, 0) :
            raise ValueError, "not enough occurrences"
        self[value] -= times
        if self[value] == 0 :
            del self[value]
    def remove (self, values, times=1) :
        """Remove values to the multiset.

        >>> m = MultiSet()
        >>> m.add([1, 2, 2, 3], 2)
        >>> list(sorted(m.items()))
        [1, 1, 2, 2, 2, 2, 3, 3]
        >>> m.remove(2, 3)
        >>> list(sorted(m.items()))
        [1, 1, 2, 3, 3]
        >>> m.remove([1, 3], 2)
        >>> list(sorted(m.items()))
        [2]

        @param values: the values to remove or a single value to remove
        @type values: any atomic object (C{str} included) or an
          iterable object
        @param times: the number of times each value should be removed
        @type times: non negative C{int}"""
        for value in iterate(values) :
            self._remove(value, times)
    def __call__ (self, value) :
        """Number of occurrences of C{value}.

        >>> m = MultiSet([1, 1, 2, 3, 3, 3])
        >>> m(1), m(2), m(3), m(4)
        (2, 1, 3, 0)

        @param value: the value the count
        @type value: C{object}
        @rtype: C{int}
        """
        return self.get(value, 0)
    def __iter__ (self) :
        """
        Iterate over the values (with repetitions).

        Use C{MultiSet.keys} to ignore repetitions.

        @return: an iterator on the elements
        @rtype: C{iterator}"""
        for value in dict.__iter__(self) :
            for count in range(self[value]) :
                yield value
    def items (self) :
        """
        Return the list of items with repetitions. The list without
        repetitions can be retrieved with the C{key} method.

        >>> m = MultiSet([1, 2, 2, 3])
        >>> list(sorted(m.items()))
        [1, 2, 2, 3]
        >>> list(sorted(m.keys()))
        [1, 2, 3]

        @return: list of items with repetitions
        @rtype: C{list}"""
        return list(iter(self))
    def __str__ (self) :
        return "{%s}" % ", ".join(str(x) for x in self)
    def __repr__ (self) :
        return "MultiSet([%s])" % ", ".join(repr(x) for x in self)
    def __len__ (self) :
        """Return the number of elements, including repetitions.

        >>> len(MultiSet([1, 2] * 3))
        6

        @rtype: C{int}
        """
        if self.size() == 0 :
            return 0
        else :
            return reduce(operator.add, self.values())
    def size (self) :
        """Return the number of elements, excluding repetitions.

        >>> MultiSet([1, 2] * 3).size()
        2

        @rtype: C{int}
        """
        return dict.__len__(self)
    def __add__ (self, other) :
        """Adds two multisets.

        >>> MultiSet([1, 2, 3]) + MultiSet([2, 3, 4])
        MultiSet([1, 2, 2, 3, 3, 4])

        @param other: the multiset to add
        @type other: C{MultiSet}
        @rtype: C{MultiSet}
        """
        result = self.copy()
        for value, times in dict.items(other) :
            result._add(value, times)
        return result
    def __sub__ (self, other) :
        """Substract two multisets.

        >>> MultiSet([1, 2, 3]) - MultiSet([2, 3])
        MultiSet([1])
        >>> MultiSet([1, 2, 3]) - MultiSet([2, 3, 4])
        Traceback (most recent call last):
        ...
        ValueError: not enough occurrences

        @param other: the multiset to substract
        @type other: C{MultiSet}
        @rtype: C{MultiSet}
        """
        result = self.copy()
        for value, times in dict.items(other) :
            result._remove(value, times)
        return result
    def __mul__ (self, other) :
        """Multiplication by a non-negative integer.

        >>> MultiSet([1, 2]) * 3
        MultiSet([1, 1, 1, 2, 2, 2])

        @param other: the integer to multiply
        @type other: non-negative C{int}
        @rtype: C{MultiSet}
        """
        if other < 0 :
            raise ValueError, "negative values are forbidden"
        elif other == 0 :
            return MultiSet()
        else :
            result = self.copy()
            for value in self.keys() :
                result[value] *= other
            return result
    def __eq__ (self, other) :
        """Test for equality.

        >>> MultiSet([1, 2, 3]*2) == MultiSet([1, 2, 3]*2)
        True
        >>> MultiSet([1, 2, 3]) == MultiSet([1, 2, 3, 3])
        False

        @param other: the multiset to compare with
        @type other: C{MultiSet}
        @rtype: C{bool}
        """
        if len(self) != len(other) :
            return False
        else :
            for val in self :
                try :
                    if self[val] != other[val] :
                        return False
                except KeyError :
                    return False
        return True
    def __ne__ (self, other) :
        """Test for difference.

        >>> MultiSet([1, 2, 3]*2) != MultiSet([1, 2, 3]*2)
        False
        >>> MultiSet([1, 2, 3]) != MultiSet([1, 2, 3, 3])
        True

        @param other: the multiset to compare with
        @type other: C{MultiSet}
        @rtype: C{bool}
        """
        return not(self == other)
    def __lt__ (self, other) :
        """Test for strict inclusion.

        >>> MultiSet([1, 2, 3]) < MultiSet([1, 2, 3, 4])
        True
        >>> MultiSet([1, 2, 3]) < MultiSet([1, 2, 3, 3])
        True
        >>> MultiSet([1, 2, 3]) < MultiSet([1, 2, 3])
        False
        >>> MultiSet([1, 2, 3]) < MultiSet([1, 2])
        False
        >>> MultiSet([1, 2, 2]) < MultiSet([1, 2, 3, 4])
        False

        @param other: the multiset to compare with
        @type other: C{MultiSet}
        @rtype: C{bool}
        """
        if not set(self.keys()) <= set(other.keys()) :
            return False
        result = False
        for value, times in dict.items(self) :
            count = other(value)
            if times > count :
                return False
            elif times < count :
                result = True
        return result or (dict.__len__(self) < dict.__len__(other))
    def __le__ (self, other) :
        """Test for inclusion inclusion.

        >>> MultiSet([1, 2, 3]) <= MultiSet([1, 2, 3, 4])
        True
        >>> MultiSet([1, 2, 3]) <= MultiSet([1, 2, 3, 3])
        True
        >>> MultiSet([1, 2, 3]) <= MultiSet([1, 2, 3])
        True
        >>> MultiSet([1, 2, 3]) <= MultiSet([1, 2])
        False
        >>> MultiSet([1, 2, 2]) <= MultiSet([1, 2, 3, 4])
        False

        @param other: the multiset to compare with
        @type other: C{MultiSet}
        @rtype: C{bool}
        """
        if not set(self.keys()) <= set(other.keys()) :
            return False
        for value, times in dict.items(self) :
            count = other(value)
            if times > count :
                return False
        return True
    def __gt__ (self, other) :
        """Test for strict inclusion.

        >>> MultiSet([1, 2, 3, 4]) > MultiSet([1, 2, 3])
        True
        >>> MultiSet([1, 2, 3, 3]) > MultiSet([1, 2, 3])
        True
        >>> MultiSet([1, 2, 3]) > MultiSet([1, 2, 3])
        False
        >>> MultiSet([1, 2]) > MultiSet([1, 2, 3])
        False
        >>> MultiSet([1, 2, 3, 4]) > MultiSet([1, 2, 2])
        False

        @param other: the multiset to compare with
        @type other: C{MultiSet}
        @rtype: C{bool}
        """
        return other.__lt__(self)
    def __ge__ (self, other) :
        """Test for inclusion.

        >>> MultiSet([1, 2, 3, 4]) >= MultiSet([1, 2, 3])
        True
        >>> MultiSet([1, 2, 3, 3]) >= MultiSet([1, 2, 3])
        True
        >>> MultiSet([1, 2, 3]) >= MultiSet([1, 2, 3])
        True
        >>> MultiSet([1, 2]) >= MultiSet([1, 2, 3])
        False
        >>> MultiSet([1, 2, 3, 4]) >= MultiSet([1, 2, 2])
        False

        @param other: the multiset to compare with
        @type other: C{MultiSet}
        @rtype: C{bool}
        """
        return other.__le__(self)

class Substitution (object) :
    """Map names to values or names, equals the identity where not defined.

    Substitutions support the C{+} operation (union with consistency
    check between the two operands) and the C{*} operation which is
    the composition of functions (C{(f*g)(x)} is C{f(g(x))}).

    Several methods (eg, C{image}) return lists instead of sets, this
    avoids the restriction of having only hashable values in a
    substitution image.
    """
    def __init__ (self, *largs, **dargs) :
        """Initialise using a dictionnary as a mapping.

        The expected arguments are any ones acceptables for
        initializing a dictionnary.
        """
        self._dict = dict(*largs, **dargs)
    @pnmldump
    def __pnmldump__ (self, doc) :
        result = doc.createElement("substitution")
        for name, value in self._dict.items() :
            item = doc.createElement("item")
            item.setAttribute("name", str(name))
            item.appendChild(obj_to_pnml(doc, value))
            result.appendChild(item)
        return result
    @pnmlload
    def __pnmlload__ (cls, node) :
        result = cls()
        for item in node.childNodes :
            name = string(item.getAttribute("name"))
            value = pnml_to_obj(item.firstChild)
            result._dict[name] = value
        return result
    def items (self) :
        "Return the list of pairs (name, value)."
        return self._dict.items()
    def domain (self) :
        "Return the set of mapped names."
        return set(self._dict.keys())
    def image (self) :
        "Return the set of values associated to the names."
        return self._dict.values()
    def __contains__ (self, name) :
        "Test if a name is mapped."
        return name in self._dict
    def __iter__ (self) :
        "Iterate over the mapped names."
        return iter(self._dict)
    def __str__ (self) :
        "Return a compact string representation."
        return "{%s}" % ", ".join(["%s -> %s" % (str(var), str(val))
                                   for var, val in self.items()])
    def __repr__ (self) :
        "Return a string representation suitable for C{eval}."
        return "%s(%s)" % (self.__class__.__name__,
                           ", ".join(("%s=%s" % (str(var), repr(val))
                                      for var, val in self.items())))
    def dict (self) :
        "Return the mapping as a dictionnary."
        return self._dict.copy()
    def copy (self) :
        "Copy the mapping."
        return Substitution(self.dict())
    def __getitem__ (self, var) :
        """Return the mapped value.

        Fails with C{DomainError} if C{var} is not mapped.

        >>> s = Substitution(x=1, y=2)
        >>> s['x']
        1
        >>> s['z']
        Traceback (most recent call last):
        ...
        DomainError: unbound variable 'z'
        """
        try :
            return self._dict[var]
        except KeyError :
            raise DomainError, "unbound variable '%s'" % str(var)
    def __call__ (self, var) :
        """Return the mapped value.

        Never fails but return C{var} if it is not mapped.

        >>> s = Substitution(x=1, y=2)
        >>> s('x')
        1
        >>> s('z')
        'z'
        """
        try :
            return self._dict[var]
        except KeyError :
            return var
    def __add__ (self, other) :
        """Add two substitution.

        Fails with C{DomainError} if the two substitutions map a same
        name to different values.

        >>> s = Substitution(x=1, y=2) + Substitution(y=2, z=3)
        >>> s('x'), s('y'), s('z')
        (1, 2, 3)
        >>> Substitution(x=1, y=2) + Substitution(y=4, z=3)
        Traceback (most recent call last):
        ...
        DomainError: conflict on 'y'
        """
        for var in self :
            if var in other and (self[var] != other[var]) :
                raise DomainError, "conflict on '%s'" % str(var)
        s = self.__class__(self.dict())
        s._dict.update(other.dict())
        return s
    def __mul__ (self, other) :
        """Compose two substitutions.

        The composition of f and g is such that (f*g)(x) = f(g(x)).

        >>> f = Substitution(a=1, d=3, y=5)
        >>> g = Substitution(b='d', c=2, e=4, y=6)
        >>> h = f*g
        >>> h('a'), h('b'), h('c'), h('d'), h('e'), h('y'), h('x')
         (1, 3, 2, 3, 4, 6, 'x')
        """
        res = self.copy()
        for var in other :
            res._dict[var] = self(other(var))
        return res
