"""A plugins system.

The first example shows how to load a plugin: we load
C{snakes.plugins.hello} and plug it into C{snakes.nets}, which results
in a new module that actually C{snakes.nets} extended by
C{snakes.plugins.hello}.

>>> import snakes.plugins as plugins
>>> hello_nets = plugins.load('hello', 'snakes.nets')
>>> n = hello_nets.PetriNet('N')
>>> n.hello()
Hello from N
>>> n = hello_nets.PetriNet('N', hello='Hi, this is %s!')
>>> n.hello()
Hi, this is N!

The next example shows how to simulate the effect of C{import module}:
we give to C{load} a thrid argument that is the name of the created
module, from which it becomes possible to import names or C{*}.

B{Warning:} this feature will not work C{load} is not called from the
module where we then do the C{from ... import ...}. This is exactly
the same when, from a module C{foo} that you load a module C{bar}: if
C{bar} loads other modules they will not be imported in C{foo}.

>>> plugins.load('hello', 'snakes.nets', 'another_version')
<module ...>
>>> from another_version import PetriNet
>>> n = PetriNet('another net')
>>> n.hello()
Hello from another net
>>> n = PetriNet('yet another net', hello='Hi, this is %s!')
>>> n.hello()
Hi, this is yet another net!

How to define a plugin is explained in the example C{hello}.
"""

import imp, sys, inspect

def _copy (old, new) :
    for attr in ('__module__', '__name__', '__doc__') :
        setattr(new, attr, getattr(old, attr))
    new.__dict__.update(old.__dict__)

def depend (*names) :
    """A decorator to handle dependencies between plugins.

    For instance, if a plugin depends ont C{egg} and C{spam}, its
    function C{extend} should be decorated by C{@depend('egg',
    'spam')}.

    This decorator takes care not to load a plugin twice but it will
    I{not} check for circular dependencies.

    @param names: plugin names that are depended on
    @type names: each is an instance of C{str}
    """
    def wrapper (fun) :
        def extend (module) :
            try :
                loaded = set(module.__plugins__)
            except :
                loaded = set()
            for name in names :
                if name not in loaded :
                    module = load(name, module)
            return fun(module)
        _copy(fun, extend)
        return extend
    return wrapper

def build (name, module, *classes) :
    """Builds an extended module.

    The parameter C{module} is exactly that taken by the function
    C{extend} of a plugin. This list argument C{classes} holds all the
    classes, constructed in C{extend}, that are extensions of classes
    from C{module}. The resulting value should be returned by
    C{extend}.

    @param name: the name of the constructed module
    @type name: C{str}
    @param module: the extended module
    @type module: C{module}
    @param classes: the sub-classes
    @type classes: each is a class object
    @return: the new module
    @rtype: C{module}
    """
    result = imp.new_module(name)
    result.__dict__.update(module.__dict__)
    result.__dict__.update((cls.__name__, cls) for cls in classes)
    result.__plugins__ = (module.__dict__.get("__plugins__",
                                              (module.__name__,))
                          + (name,))
    for cls in classes :
        cls.__plugins__ = result.__plugins__
    return result

def load (plugins, base, name=None) :
    """Load plugins.

    C{plugins} can be a single plugin name or module or a list of such
    values. If C{name} is not C{None}, the extended module is loaded
    ad C{name} in C{sys.modules} as well as in the global environment
    from which C{load} was called.

    @param plugins: the module that implements the plugin, or its name,
      or a collection of such values
    @type plugins: C{str} or C{module}, or a C{list}/C{tuple}/... of
      such values
    @param base: the module being extended or its name
    @type base: C{str} or C{module}
    @param name: the name of the created module
    @type name: C{str}
    @return: the extended module
    @rtype: C{module}
    """
    if type(base) is str :
        result = __import__(base, None, None, ["__name__"])
    else :
        result = base
    if isinstance(plugins, str) :
        plugins = [plugins]
    else :
        try :
            plugins = list(plugins)
        except TypeError :
            plugins = [plugins]
    for i, p in enumerate(plugins) :
        if isinstance(p, str) and not p.startswith("snakes.plugins.") :
            plugins[i] = "snakes.plugins." + p
    for plug in plugins :
        if type(plug) is str :
            plug = __import__(plug, None, None, ["__name__"])
        result = plug.extend(result)
    if name is not None :
        result.__name__ = name
        sys.modules[name] = result
    return result

def extension (base) :
    def wrap (fun) :
        def extend (module) :
            classes = fun(module)
            if type(classes) is not tuple :
                classes = (classes,)
            return build(fun.__module__, module, *classes)
        module = sys.modules[fun.__module__]
        module.__test__ = {"extend" : extend}
        classes = fun(__import__(base, None, None, ["__name__"]))
        if type(classes) is not tuple :
            classes = (classes,)
        for cls in  classes :
            setattr(module, cls.__name__, cls)
        _copy(fun, extend)
        return extend
    return wrap
