"""Export nets to the format of external tools.

Currently supported formats are: Maria, Helena and PEP. PNML is also
supported as it is builtin SNAKES.

>>> import snakes.plugins, snakes.typing
>>> nets = snakes.plugins.load('export', 'snakes.nets', 'nets')
>>> from nets import *
>>> n = PetriNet('N')
>>> n.add_place(Place('p', [0], snakes.typing.tInteger))
>>> n.add_transition(Transition('t', Expression('x<10000')))
>>> n.add_input('p', 't', Variable('x'))
>>> n.add_output('p', 't', Expression('x+1'))

>>> n.export('test-export.pnml', 'pnml')
>>> print open('test-export.pnml').read()
    <?xml version="1.0" ?>
    <pnml>
      <net class="PetriNet" id="N" plugins="snakes.nets:snakes.plugins.export">
        <place class="Place" id="p" plugins="snakes.nets" type="Instance(int)">
          <marking>
            <object class="int" method="direct" value="0"/>
          </marking>
        </place>
        <transition class="Transition" id="t" plugins="snakes.nets">
          <guard>
            <annotation class="Expression" expr="x&lt;10000" plugins="snakes.nets"/>
          </guard>
        </transition>
        <arc source="p" target="t">
          <annotation class="Variable" name="x" plugins="snakes.nets"/>
        </arc>
        <arc source="t" target="p">
          <annotation class="Expression" expr="x+1" plugins="snakes.nets"/>
        </arc>
      </net>
    </pnml>

>>> n.export('test-export.maria', 'maria')
>>> print open('test-export.maria').read()

>>> n.export('test-export.helena', 'helena')
>>> print open('test-export.helena').read()

>>> n.export('test-export.pep', 'pep')
>>> print open('test-export.pep').read()

"""

import snakes.plugins, snakes.pnml, snakes.typing

@snakes.plugins.extension("snakes.nets")
def extend (module) :
    class PetriNet (module.PetriNet) :
        def export (self, filename, format) :
            try :
                handler = getattr(self, "_export_" + format.lower())
            except :
                raise ValueError, "unknown format '%s'" % format
            outfile = open(filename, "w")
            handler(outfile)
            outfile.close()
        def _export_maria (self, outfile) :
            for place in self.place() :
                t = place.checker()
                if isinstance(t, snakes.typing.OneOf) :
                    print >>outfile, ("typedef enum {%s} type_%s;"
                                      % (", ".join((str(v) for v in t)),
                                         place.name))
                elif isinstance(t, snakes.typing.Range) and t._step == 1 :
                    print >>outfile, ("typedef int (%s..%s) type_%s;"
                                      % (str(t._first), str(t._last),
                                         place.name))
                elif t is snakes.typing.tInteger :
                    print >>outfile, ("typedef int type_%s;" % place.name)
                else :
                    raise ValueError, "unsupported place type '%s'" % str(t)
                if place.is_empty() :
                    print >>outfile, "place %s type_%s;" % (place.name,
                                                            place.name)
                else :
                    print >>outfile, ("place %s type_%s : %s;"
                                      % (place.name, place.name,
                                         ", ".join((str(v) for v in
                                                    place.tokens))))
            for trans in self.transition() :
                print >>outfile, "trans", trans.name
                print >>outfile, "  in {"
        def _export_helena (self, outfile) :
            pass
        def _export_pep (self, outfile) :
            pass
        def _export_pnml (self, outfile) :
            outfile.write(snakes.pnml.dumps(self))
    return PetriNet
