"""A plugin to use external integer values modelled by Lash.

>>> import snakes.plugins
>>> snakes.plugins.load('lashdata', 'snakes.nets', 'nets')
<module 'nets' from ...>
>>> from nets import *
>>> from snakes.data import Substitution
>>> import lash
>>> n = PetriNet('N', lash=Data(a=1, b=0))
>>> n.add_transition(Transition('t'), condition='a<3', update='a=a+1; b=b+2')
>>> n.lash.variables()
('a', 'b')
>>> n.lash.state == lash.NDD.new_vector(n.lash.base, (1, 0))
True
>>> n.transition('t').fire(Substitution())
>>> n.lash.state == lash.NDD.new_vector(n.lash.base, (2, 2))
True
>>> n.transition('t').fire(Substitution())
>>> n.lash.state == lash.NDD.new_vector(n.lash.base, (3, 4))
True
>>> n.transition('t').fire(Substitution())
Traceback (most recent call last):
  ...
ValueError: transition not enabled for {}
>>> n.lash.reset(a=1, b=0)
>>> n.lash.state == lash.NDD.new_vector(n.lash.base, (1, 0))
True
>>> n.transition('t').firestar(Substitution())
True
>>> n.lash.state == (lash.NDD.new_vector(n.lash.base, (2, 2))
...                  | lash.NDD.new_vector(n.lash.base, (3, 4)))
True
"""

import snakes.plugins, snakes.compyler
from snakes.data import cross, MultiSet # FIXME: why is MultiSet needed for doctest?
import lash

class _LLC (object) :
    "list of _LC"
    def __init__ (self, lst) :
        self._lst = list(lst)
    def __ne__ (self, other) :
        raise ValueError, "not a linear expression"
    def __and__ (self, other) :
        if isinstance(other, _LC) :
            return self.__class__(self._lst + [other])
        else :
            return self.__class__(self._lst + other._lst)
    def __repr__ (self) :
        return "%s([%s])" % (self.__class__.__name__,
                             ", ".join(repr(x) for x in self._lst))
    def __call__ (self, num) :
        """
        >>> _LLC([_LC(x=1, _=0), _LC(x=-1, _=0)])(dict(x=0, y=1))
        ([[1, 0], [-1, 0]], [0, 0])
        """
        p, q = [], []
        for coefs in self._lst :
            x, y = coefs(num)
            p.extend(x), q.extend(y)
        return p, q

class _LC (dict) :
    "linear combination"
    def copy (self) :
        """
        >>> d = _LC(a=1)
        >>> type(d.copy())
        <class 'snakes.plugins.lashdata._LC'>
        """
        return self.__class__(self)
    def __repr__(self) :
        return "%s(%s)" % (self.__class__.__name__,
                           ", ".join("%s=%s" % (name, value) for name, value in self.items()))
    def __add__ (self, other) :
        """
        >>> _LC(a=1, b=2) + _LC(b=3, c=4)
        _LC(a=1, c=4, b=5)
        >>> _LC(a=1, b=2) + _LC(_=5)
        _LC(a=1, b=2, _=5)
        """
        result = self.copy()
        for name, value in other.items() :
            result[name] = result.get(name, 0) + value
        return result
    def __sub__ (self, other) :
        """
        >>> _LC(a=1, b=2) - _LC(b=3, c=4)
        _LC(a=1, c=-4, b=-1)
        >>> _LC(a=1, b=2) - _LC(_=5)
        _LC(a=1, b=2, _=-5)
        >>> _LC(_=5) - _LC(a=1, b=2)
        _LC(a=-1, b=-2, _=5)
        """
        result = self.copy()
        for name, value in other.items() :
            result[name] = result.get(name, 0) - value
        return result
    def _const (self) :
        return ("_" in self) and (len(self) == 1)
    def __mul__ (self, other) :
        """
        >>> _LC(a=1) * _LC(_=3)
        _LC(a=3)
        >>> _LC(a=1) * _LC(b=1)
        Traceback (most recent call last):
          ...
        ValueError: not a linear expression
        """
        if self._const() :
            result = other.copy()
            num = self["_"]
        elif other._const() :
            result = self.copy()
            num = other["_"]
        else :
            raise ValueError, "not a linear expression"
        for name in result :
            result[name] *= num
        return result
    def __neg__ (self) :
        """
        >>> -(_LC(a=1))
        _LC(a=-1)
        """
        return self.__class__((name, -value) for name, value in self.items())
    def __pos__ (self) :
        return self.copy()
    def __le__ (self, other) :
        return self - other
    def __ge__ (self, other) :
        return (-self).__le__(-other)
    def __lt__ (self, other) :
        return self.__le__(other - self.__class__(_=1))
    def __gt__ (self, other) :
        return (-self).__le__(-other - self.__class__(_=1))
    def __eq__ (self, other) :
        """
        >>> _LC(a=1) == _LC(_=2)
        _LLC([_LC(a=1, _=-2), _LC(a=-1, _=2)])
        """
        return (self <= other) & (other <= self)
    def __ne__ (self, other) :
        raise ValueError, "not a linear expression"
    def __and__ (self, other) :
        """
        >>> (_LC(x=1) <= _LC(_=0)) & (_LC(x=1) >= _LC(_=0))
        _LLC([_LC(x=1, _=0), _LC(x=-1, _=0)])
        """
        if not isinstance(other, _LC) :
            return other.__and__(self)
        else :
            return _LLC([self, other])
    def __call__ (self, num) :
        """
        >>> (_LC(a=1)+_LC(_=2)*_LC(b=1)>_LC(c=1)+_LC(_=2))(dict(a=0, b=1, c=2))
        ([[-1, -2, 1]], [-3])
        >>> (_LC(a=1) == _LC(_=2))(dict(a=0, b=1, c=2))
        ([[1, 0, 0], [-1, 0, 0]], [2, -2])
        """
        line = [0] * len(num)
        for name in self :
            if name != "_" :
                line[num[name]] = self[name]
        return [line], [-self.get("_", 0)]

class _Rename (snakes.compyler.Parser) :
    """
    >>> l = _Rename()
    >>> l('a + 2*b > c + 2')
    '((_LC(a=1)) + ((_LC(_=2)) * (_LC(b=1)))) > ((_LC(c=1)) + (_LC(_=2)))'
    >>> eval(l('a + 2*b > c + 2'))
    _LC(a=-1, c=1, b=-2, _=3)
    >>> l('a>1 and b<2')
    '((_LC(a=1)) > (_LC(_=1))) & ((_LC(b=1)) < (_LC(_=2)))'
    """
    def __init__ (self) :
        self._unsupported = self._unsupported.copy()
        for name in ["AssName", "AssTuple", "Backquote",
                     "Bitxor", "CallFunc", "Dict", "Div", "Getattr", "Invert",
                     "Keyword", "LeftShift", "List", "ListComp",
                     "ListCompFor", "ListCompIf", "Mod",
                     "Not", "Or", "Power", "RightShift", "Slice",
                     "Sliceobj", "Subscript", "Tuple"] :
            self._unsupported[name] = getattr(self, "_%s_" % name).__doc__
    def _And_ (self, node) :
        return " & ".join("(%s)" % self[n] for n in node.nodes)
    def _Name_ (self, node) :
        if node.name == "_" :
            raise ValueError, "forbidden variable name '_'"
        return "_LC(%s=1)" % node.name
    def _Const_ (self, node) :
        if not isinstance(node.value, int) :
            raise ValueError, "forbidden value: %s" % repr(node.value)
        return "_LC(_=%s)" % repr(node.value)

class _Update (_Rename) :
    """
    >>> l = _Update()
    >>> l('a=1; b=b+1')
    {'a': '_LC(_=1)', 'b': '(_LC(b=1)) + (_LC(_=1))'}
    """
    def __init__ (self) :
        self._unsupported = self._unsupported.copy()
        for name in ["And", "AssTuple", "Backquote",
                     "Bitxor", "CallFunc", "Dict", "Div", "Getattr", "Invert",
                     "Keyword", "LeftShift", "List", "ListComp",
                     "ListCompFor", "ListCompIf", "Mod",
                     "Not", "Or", "Power", "RightShift", "Slice",
                     "Sliceobj", "Subscript", "Tuple", "Compare"] :
            self._unsupported[name] = getattr(self, "_%s_" % name).__doc__
        del self._unsupported["Assign"]
    def __call__ (self, expr) :
        self._result = {}
        _Rename.__call__(self, expr)
        return self._result
    def _Name_ (self, node) :
        if node.name == "_" :
            raise ValueError, "forbidden variable name '_'"
        return "_LC(%s=1)" % node.name
    def _AssName_ (self, node) :
        if node.name == "_" :
            raise ValueError, "forbidden variable name '_'"
        return node.name
    def _Const_ (self, node) :
        if not isinstance(node.value, int) :
            raise ValueError, "forbidden value: %s" % repr(node.value)
        return "_LC(_=%s)" % repr(node.value)
    def _Stmt_ (self, node) :
        "statement"
        return "\n".join(self[n] for n in node.nodes)
    def _Assign_ (self, node) :
        name = node.nodes[0].name
        if name in self._result :
            raise ValueError, "'%s' assigned twice" % name
        self._result[name] = self[node.expr]
        return str(node)

class Data (object) :
    base = 2
    _rename = _Rename()
    _update = _Update()
    def __init__ (self, state=None, *variables, **vector) :
        """
        >>> d = Data(x=0, y=1)
        >>> d.contains(x=0, y=1)
        True
        >>> d.contains(x=1, y=1)
        False
        >>> d = Data(lash.NDD.new_eq2(Data.base), 'x', 'y')
        >>> d.contains(x=1, y=1)
        True
        >>> d.contains(x=0, y=1)
        False
        """
        if state is not None :
            if len(variables) < len(state) :
                raise ValueError, "not enough variables"
            elif len(variables) > len(state) :
                raise ValueError, "too much variables"
            self._num = dict((v, i) for i, v in enumerate(variables))
            self.state = state.copy()
        else :
            self._num = {}
            val = [0] * len(vector)
            for i, v in enumerate(vector) :
                self._num[v] = i
                val[i] = int(vector[v])
            self.state = lash.NDD.new_vector(self.base, val)
        self._info = {}
        self._lt = {}
    def copy (self) :
        result = self.__class__()
        result._num = self._num.copy()
        result.state = self.state.copy()
        result._info = {}
        for trans, info in self._info.iteritems() :
            result._info[trans] = info.copy()
        result._lt = {}
        for trans, lt in self._lt.iteritems() :
            result._lt[trans] = lt.copy()
        return result
    def contains (self, **vector) :
        val = [0] * len(self._num)
        for v in vector :
            val[self._num[v]] = int(vector[v])
        return val in self.state
    def __len__ (self) :
        return len(self._num)
    def reset (self, **values) :
        val = [0] * len(self)
        for v in values :
            val[self._num[v]] = int(values[v])
        self.state = lash.NDD.new_vector(self.base, val)
    def variables (self) :
        result = [0] * len(self)
        for name, num in self._num.items() :
            result[num] = name
        return tuple(result)
    def linear_transf (self, guard, update) :
        """
        >>> d = Data(x=1, y=2)
        >>> d.variables()
        ('y', 'x')
        >>> lt = d.linear_transf('y==0', 'x=x+1; y=y-2')
        >>> lt(d.state) == lash.NDD.new_empty(d.base, 2)
        True
        >>> lt = d.linear_transf('y==2', 'x=x+1; y=y-2')
        >>> lt(d.state) == lash.NDD.new_vector(d.base, (0, 2))
        True
        >>> lt = d.linear_transf(True, 'x=y; y=x')
        >>> lt(d.state) == lash.NDD.new_vector(d.base, (1, 2))
        True
        >>> d.linear_transf('x!=1', None)
        Traceback (most recent call last):
          ...
        ValueError: not a linear expression
        >>> lt = d.linear_transf(True, None)
        >>> lt(lash.NDD.new_eq2(d.base)) == lash.NDD.new_eq2(d.base)
        True
        >>> lt = d.linear_transf(False, None)
        >>> lt(lash.NDD.new_eq2(d.base)) == lash.NDD.new_empty(d.base, 2)
        True
        """
        if guard is True :
            p, q = self._expr("0")
        elif guard is False :
            p, q = self._expr("1")
        else :
            p, q = self._expr(guard)
        a, b = self._assign(update)
        return lash.LinearTransf.new(a, b, p, q)
    def zone (self, constraint) :
        """
        >>> d = Data(x=1, y=2)
        >>> d.variables()
        ('y', 'x')
        >>> d.zone('x+y<=1') == lash.NDD.new_inequation(d.base, (1, 1), 1)
        True
        >>> d.zone('x==y') == lash.NDD.new_eq2(d.base)
        True
        >>> d.zone('(x==0) and (y==0)') == lash.NDD.new_vector(d.base, (0, 0))
        True
        """
        a, b = self._assign(None)
        zn = lash.NDD.new_zn(self.base, len(self))
        result = lash.NDD.new_empty(self.base, len(self))
        z = self._expr(constraint)
        if isinstance(z, tuple) :
            p, q = z
            result = lash.LinearTransf.new(a, b, p, q)(zn)
        else :
            for p, q in z :
                result = result | lash.LinearTransf.new(a, b, p, q)(zn)
        return result
    def add_transition (self, trans, condition, update) :
        name = trans.name
        self._lt[name] = self.linear_transf(condition, update)
        self._info[name] = lash.LinearTransfInfo(self._lt[name])
        trans.lash = self
    def enabled (self, trans) :
        return not self._info[trans.name](self.state).is_empty()
    def fire (self, trans) :
        self.state = self._info[trans.name](self.state)
    def firestar (self, trans) :
        self.fire(trans)
        self.state = self._info[trans.name].star(self.state)
    def _expr (self, expression) :
        """
        >>> d = Data(a=0, b=0, c=0)
        >>> d.variables()
        ('a', 'c', 'b')
        >>> d._expr('(a + 2*b > c + 2) and (a+c+b<10)')
        ([[-1, 1, -2],
          [1, 1, 1]], [-3, 9])
        >>> d._expr('3')
        ([[0, 0, 0]], [-3])
        """
        return eval(self._rename(expression))(self._num)
    def _assign (self, assignments) :
        """
        >>> d = Data(a=0, b=0, c=0)
        >>> d.variables()
        ('a', 'c', 'b')
        >>> d._assign('a=b+1; b=c+b-1; c=b')
        ([[0, 0, 1],
          [0, 0, 1],
          [0, 1, 1]], [1, 0, -1])
        >>> d._assign('a=1; b=2; c=3')
        ([[0, 0, 0],
          [0, 0, 0],
          [0, 0, 0]], [1, 3, 2])
        """
        a = self._identity()
        b = [0] * len(self)
        if assignments is None :
            return a, b
        for target, expr in self._update(assignments).items() :
            num = self._num[target]
            line, const = eval(expr)(self._num)
            a[num] = line[0]
            b[num] = - const[0]
        return a, b
    def _identity (self) :
        dimen = len(self)
        result = []
        for i in range(dimen) :
            line = [0] * dimen
            line[i] = 1
            result.append(line)
        return result

@snakes.plugins.extension("snakes.nets")
def extend (module) :
    "Build the extended module"
    class Marking (module.Marking) :
        def __eq__ (self, other) :
            if hasattr(self, "lash") and hasattr(other, "lash") :
                return self.same(other) and self.lash == other.lash
            elif hasattr(self, "lash") or hasattr(other, "lash") :
                raise ValueError, "markings not compatible"
            else :
                return self.same(other)
        def same (self, other) :
            return module.Marking.__eq__(self, other)
        def copy (self, **args) :
            _lash = args.pop("lash", True)
            result = module.Marking.copy(self, **args)
            if _lash :
                try :
                    result.lash = self.lash.copy()
                except AttributeError :
                    pass
            return result
    class PetriNet (module.PetriNet) :
        def __init__ (self, name, **args) :
            _lash = args.pop("lash", None)
            module.PetriNet.__init__(self, name, **args)
            if _lash is None :
                _lash = Data(x=0)
            self.lash = _lash
        def copy (self) :
            result = module.PetriNet.copy(self)
            result.lash.__dict__.update(self.lash.copy().__dict__)
            return result
        def add_transition (self, trans, **args) :
            condition = args.pop("condition", True)
            update =  args.pop("update", None)
            module.PetriNet.add_transition(self, trans, **args)
            self.lash.add_transition(trans, condition, update)
        def get_marking (self, **args) :
            result = Marking(module.PetriNet.get_marking(self, **args))
            result.lash = self.lash.state.copy()
            return result
        def set_marking (self, marking, **args) :
            module.PetriNet.set_marking(self, marking, **args)
            try :
                self.lash.state = marking.lash.copy()
            except AttributeError :
                pass
    class Transition (module.Transition) :
        def _check (self, binding, tokens, input) :
            return (module.Transition._check(self, binding, tokens, input)
                    and self.lash.enabled(self))
        def fire (self, binding, **args) :
            module.Transition.fire(self, binding, **args)
            self.lash.fire(self)
        def firestar (self, binding) :
            removed, created = self.flow(binding)
            if removed != created :
                self.fire(binding)
                return False
            else :
                module.Transition.fire(self, binding)
                self.lash.firestar(self)
                return True
    class StateGraph (module.StateGraph) :
        def __init__ (self, net, **args) :
            self._loops = args.pop("loops", False)
            self._fold = args.pop("fold", False)
            self._cycles = args.pop("cycles", False)
            self._remove = args.pop("remove", False)
            if self._fold :
                self._remove = True
            if self._remove :
                self._cycles = True
            if self._cycles :
                self._loops = True
            self._data = {}
            module.StateGraph.__init__(self, net, **args)
        def goto (self, state, **args) :
            module.StateGraph.goto(self, state, **args)
            #FIXME: find why this is necessary
            #state = self.net.lash.state.copy()
            self.net.lash.state = self._marking[self._current].lash.copy()
            #if state != self.net.lash.state.copy() :
            #    raise RuntimeError, " ".join(str(x) for x in [self._current, self._done, self._todo, self._removed])
        def _get_state (self, marking) :
            stripped = marking.copy(lash=False)
            if stripped not in self._data :
                return None
            for state, data in self._data[stripped].iteritems() :
                if marking.lash.issubset(data) :
                    return state
            return None
        def _create_state (self, marking, source, trans, mode) :
            stripped = marking.copy(lash=False)
            if stripped not in self._data :
                state = module.StateGraph._create_state(self, marking, source, trans, mode)
                self._data[stripped] = {state: marking.lash.copy()}
                return state
            ancestors = []
            if self._fold and marking.same(self[source]) :
                ancestors.append(source)
                marking.lash.update(self._data[stripped][source])
            elif self._cycles or self._remove :
                for state, data in self._data[stripped].iteritems() :
                    if data.issubset(marking.lash) :
                        ancestors.append(state)
                        if len(ancestors) > 0 :
                            self._saturate_cycles(ancestors, source,
                                                  trans, marking)
            state = module.StateGraph._create_state(self, marking, source,
                                                    trans, mode)
            self._data[stripped][state] = marking.lash.copy()
            if len(ancestors) > 0 :
                self._remove_covered(ancestors, state)
            return state
        def _saturate_cycles (self, sources, target, trans, marking) :
            lt = self.net.lash.linear_transf(True, None)
            for source in sources :
                for path in self._paths(source, target) :
                    #print "# cycle", path, "+", trans
                    path = list(path)
                    x = path.pop(0)
                    while len(path) > 0 :
                        t, l = path.pop(0)
                        y = path.pop(0)
                        lt = lt & self.net.lash._lt[t.name]
                        x = y
                    lt = lt & self.net.lash._lt[trans.name]
            marking.lash = lt.star(marking.lash)
        def _remove_covered (self, states, target) :
            todo = states[:]
            removed = set([])
            while len(todo) > 0 :
                state = todo.pop(0)
                if state in removed :
                    continue
                removed.add(state)
                for succ in self._succ[state] :
                    preds = set(self._pred[succ].keys())
                    preds.discard(succ)
                    if preds.issubset(removed) :
                        todo.append(succ)
            for state in removed :
                #print "# remove", state, "=>", target
                for pred in self._pred[state].keys() :
                    if pred not in removed :
                        #print "  # relink", pred, "->", target
                        for pred in self._pred[state].keys() :
                            for label in self._pred[state][pred].copy() :
                                self._create_edge(pred, target, label)
                                self._remove_edge(pred, state, label)
                self._remove_state(state)
        def _paths (self, source, target, visited=set([])) :
            for step in self._succ[source] :
                if step == source :
                    continue
                if step == target :
                    for label in self._succ[source][step] :
                        yield (source, label, target)
                if step not in visited :
                    for cycle in self._paths(step, target,
                                             visited | set([source])) :
                        for label in self._succ[source][step] :
                            yield (source, label) + cycle
        def _remove_state (self, state) :
            marking = module.StateGraph._remove_state(self, state)
            data = marking.lash
            del marking.lash
            del self._data[marking][state]
            marking.lash = data
            return marking
        def _fire (self, trans, mode) :
            "Fire trans with the given mode"
            if self._loops :
                trans.firestar(mode)
            else :
                trans.fire(mode)
    return PetriNet, Transition, Marking, StateGraph
