"""A module to save and load objects in PNML.

Petri nets objects are saved in PNML, other Python objects are saved
in a readable format when possible and pickled as a last solution.
This should result in a complete PNML serialization of any object.

Here is a list of examples that dump and rebuild object. First, basic
Petri net elements:

>>> from snakes.nets import *
>>> s = dumps(Value(1))
>>> print s
<?xml version="1.0" ?>
<pnml>
  <annotation class="Value" plugins="snakes.nets">
    <object class="int" method="direct" value="1"/>
  </annotation>
</pnml>
>>> loads(s)
Value(1)
>>> s = dumps(Variable('x'))
>>> print s
<?xml version="1.0" ?>
<pnml>
  <annotation class="Variable" name="x" plugins="snakes.nets"/>
</pnml>
>>> loads(s)
Variable('x')
>>> s = dumps(Expression('x+1'))
>>> print s
<?xml version="1.0" ?>
<pnml>
  <annotation class="Expression" expr="x+1" plugins="snakes.nets"/>
</pnml>
>>> loads(s)
Expression('x+1')
>>> s = dumps(MultiArc([Value(1), Variable('x'), Expression('x+1')]))
>>> print s
<?xml version="1.0" ?>
<pnml>
  <annotation class="MultiArc" plugins="snakes.nets">
    <annotation class="Value" plugins="snakes.nets">
      <object class="int" method="direct" value="1"/>
    </annotation>
    <annotation class="Variable" name="x" plugins="snakes.nets"/>
    <annotation class="Expression" expr="x+1" plugins="snakes.nets"/>
  </annotation>
</pnml>
>>> loads(s)
MultiArc((Value(1), Variable('x'), Expression('x+1')))
>>> s = dumps(Test(Value(1)))
>>> print s
<?xml version="1.0" ?>
<pnml>
  <annotation class="Test" plugins="snakes.nets">
    <annotation class="Value" plugins="snakes.nets">
      <object class="int" method="direct" value="1"/>
    </annotation>
  </annotation>
</pnml>
>>> loads(s)
Test(Value(1))
>>> s = dumps(Place('p', range(3), typing.tInteger))
>>> print s
<?xml version="1.0" ?>
<pnml>
  <place class="Place" id="p" plugins="snakes.nets" type="Instance(int)">
    <marking>
      <object class="int" method="direct" value="0"/>
      <object class="int" method="direct" value="1"/>
      <object class="int" method="direct" value="2"/>
    </marking>
  </place>
</pnml>
>>> loads(s)
Place('p', MultiSet([0, 1, 2]), Instance(int))
>>> s = dumps(Transition('t', Expression('x==1')))
>>> print s
<?xml version="1.0" ?>
<pnml>
  <transition class="Transition" id="t" plugins="snakes.nets">
    <guard>
      <annotation class="Expression" expr="x==1" plugins="snakes.nets"/>
    </guard>
  </transition>
</pnml>
>>> loads(s)
Transition('t', Expression('x==1'))
>>> n = PetriNet('N')
>>> n.add_place(Place('p', [0]))
>>> n.add_transition(Transition('t', Expression('x<5')))
>>> n.add_input('p', 't', Variable('x'))
>>> n.add_output('p', 't', Expression('x+1'))
>>> s = dumps(n)
>>> print s
<?xml version="1.0" ?>
<pnml>
  <net class="PetriNet" id="N" plugins="snakes.nets">
    <place class="Place" id="p" plugins="snakes.nets" type="tAll">
      <marking>
        <object class="int" method="direct" value="0"/>
      </marking>
    </place>
    <transition class="Transition" id="t" plugins="snakes.nets">
      <guard>
        <annotation class="Expression" expr="x&lt;5" plugins="snakes.nets"/>
      </guard>
    </transition>
    <arc source="p" target="t">
      <annotation class="Variable" name="x" plugins="snakes.nets"/>
    </arc>
    <arc source="t" target="p">
      <annotation class="Expression" expr="x+1" plugins="snakes.nets"/>
    </arc>
  </net>
</pnml>
>>> n = loads(s)
>>> n
PetriNet('N')
>>> n.node()
[Place('p', MultiSet([0]), tAll), Transition('t', Expression('x<5'))]
>>> n.transition('t').input()
[(Place('p', MultiSet([0]), tAll), Variable('x'))]
>>> n.transition('t').output()
[(Place('p', MultiSet([0]), tAll), Expression('x+1'))]
>>> s = dumps(n.get_marking())
>>> print s
<?xml version="1.0" ?>
<pnml>
  <marking class="Marking" plugins="snakes.nets">
    <place id="p">
      <multiset class="MultiSet" plugins="snakes.data">
        <item multiplicity="1">
          <object class="int" method="direct" value="0"/>
        </item>
      </multiset>
    </place>
  </marking>
</pnml>
>>> loads(s)
Marking({'p': MultiSet([0])})

Substitutions may be serialized also:

>>> from snakes.data import MultiSet, Substitution
>>> s = dumps(Substitution(x=1, y='azerty', z=MultiSet(range(2) + range(3) + range(4))))
>>> print s
<?xml version="1.0" ?>
<pnml>
  <substitution class="Substitution" plugins="snakes.data">
    <item name="y">
      <object class="str" method="direct" value="'azerty'"/>
    </item>
    <item name="x">
      <object class="int" method="direct" value="1"/>
    </item>
    <item name="z">
      <multiset class="MultiSet" plugins="snakes.data">
        <item multiplicity="3">
          <object class="int" method="direct" value="0"/>
        </item>
       <item multiplicity="3">
         <object class="int" method="direct" value="1"/>
        </item>
        <item multiplicity="2">
          <object class="int" method="direct" value="2"/>
        </item>
        <item multiplicity="1">
          <object class="int" method="direct" value="3"/>
        </item>
      </multiset>
    </item>
  </substitution>
</pnml>
>>> loads(s)
Substitution(y='azerty', x=1, z=MultiSet([0, 0, 0, 1, 1, 1, 2, 2, 3]))

Finally, any object can be serialized, which may result in picklinkg it.

>>> s = dumps(dict(a=4, b=Expression('x+1')))
>>> print s
<?xml version="1.0" ?>
<pnml>
  <object class="dict" method="dict">
    <item>
      <key>
        <object class="str" method="direct" value="'a'"/>
      </key>
      <value>
        <object class="int" method="direct" value="4"/>
      </value>
    </item>
    <item>
      <key>
        <object class="str" method="direct" value="'b'"/>
      </key>
      <value>
        <annotation class="Expression" expr="x+1" plugins="snakes.nets"/>
      </value>
    </item>
  </object>
</pnml>
>>> loads(s)
{'a': 4, 'b': Expression('x+1')}
>>> s = dumps(str)
>>> print s
<?xml version="1.0" ?>
<pnml>
  <object method="pickle" value="'c__builtin__\\nstr\\np1\\n.'"/>
</pnml>
>>> loads(s)
<type 'str'>
"""

import new
import xml.dom.minidom
import cPickle as pickle
import snakes.plugins

def clone (obj, cls, **args) :
    """Return copy of C{obj} having the class C{cls}

    The result is a new instance of C{cls}, having all the attributes
    and methods from C{obj} plus those taken from C{args}.

    @param obj: the object to clone
    @type obj: any
    @param cls: the class of the new instance
    @type cls: C{type}
    @return: an instance of C{cls} copied from C{obj}
    @rtype: C{cls}
    """
    result = cls.__new__(cls)
    result.__dict__.update(obj.__dict__)
    result.__dict__.update(args)
    return result

def string (text) :
    """Conversion of C{unicode} to C{str}

    >>> string('hello')
    'hello'
    >>> string(u'world!')
    'world!'

    @param text: a C{unicode} string or a C{str}
    @type text: C{unicode} or C{str}
    @return: C{text} converted to C{str}
    @rtype: C{str}
    """
    try :
        return text.encode()
    except :
        return str(text)

def dump (obj) :
    """Convert an object to a DOM tree.

    >>> from snakes.nets import Value
    >>> dump([Value(1), 1, 'string', (1, 2, 3)])
    <xml.dom.minidom.Document instance at ...>

    @param obj: the object to dump
    @type obj: any type
    @return: the DOM tree that serialize C{obj}
    @rtype: C{xml.dom.minidom.Document}
    """
    impl = xml.dom.minidom.getDOMImplementation()
    doc = impl.createDocument(None, "pnml", None)
    doc.documentElement.appendChild(obj_to_pnml(doc, obj))
    return doc

def dumps (obj) :
    """Convert an object to an XML string.

    >>> from snakes.nets import Value
    >>> print dumps([Value(1), 1, 'string', (1, 2, 3)])
    <?xml version="1.0" ?>
    <pnml>
      <object class="list" method="collection">
        <annotation class="Value" plugins="snakes.nets">
          <object class="int" method="direct" value="1"/>
        </annotation>
        <object class="int" method="direct" value="1"/>
        <object class="str" method="direct" value="'string'"/>
        <object class="tuple" method="collection">
          <object class="int" method="direct" value="1"/>
          <object class="int" method="direct" value="2"/>
          <object class="int" method="direct" value="3"/>
        </object>
      </object>
    </pnml>

    @param obj: the object to dump
    @type obj: any type
    @return: the XML string that serialize C{obj}
    @rtype: C{str}
    """
    return dump(obj).toprettyxml(indent="  ").strip()

def pnmldump (method) :
    """A decorator for methods C{__pnmldump__}

    It ensures that the returned DOM element has attributes 'class'
    and 'plugins' that allow to rebuilt it correctly.

    >>> import snakes.plugins
    >>> graphviz = snakes.plugins.load('snakes.plugins.graphviz', 'snakes.nets')
    >>> n = graphviz.PetriNet('N')
    >>> n.add_transition(graphviz.Transition('t', pos=(1, 2)))
    >>> s = dumps(n)
    >>> print s
    <?xml version="1.0" ?>
    <pnml>
      <net class="PetriNet" id="N" plugins="snakes.nets:snakes.plugins.pos:snakes.plugins.graphviz">
        <transition class="Transition" id="t" plugins="snakes.nets:snakes.plugins.pos">
          <guard>
            <annotation class="Expression" expr="True" plugins="snakes.nets"/>
          </guard>
          <graphics>
            <position x="1" y="2"/>
          </graphics>
        </transition>
      </net>
    </pnml>
    >>> r = loads(s)
    >>> t = r.node('t')
    >>> t.__class__
    <class 'snakes.plugins.pos.Transition'>
    >>> t.__plugins__
    ('snakes.nets', 'snakes.plugins.pos')
    >>> t.pos
    Position(1, 2)
    >>> r.__class__
    <class 'snakes.plugins.graphviz.PetriNet'>
    >>> r.__plugins__
    ('snakes.nets', 'snakes.plugins.pos', 'snakes.plugins.graphviz')
    >>> hasattr(r, 'draw')
    True

    @param method: the method to decorate
    @type method: instance method
    @return: a wrapper
    @rtype: instance method
    """
    def wrapper (self, doc) :
        node = method(self, doc)
        node.setAttribute("class", self.__class__.__name__)
        try :
            node.setAttribute("plugins", ":".join(self.__plugins__))
        except AttributeError :
            node.setAttribute("plugins", self.__module__)
        return node
    wrapper.__doc__ = method.__doc__
    wrapper.__name__ = method.__name__
    return wrapper

def pnmlload (method) :
    """A decorator for methods C{__pnmlload__}

    It ensures that the returned DOM element are rebuilt with the
    correct class (taking all plugins into account).
    """
    def wrapper (cls, node) :
        if hasattr(cls, "__plugins__") :
            result = clone(method(cls, node), cls)
        else :
            result = method(cls, node)
        return result
    wrapper.__doc__ = method.__doc__
    wrapper.__name__ = method.__name__
    return classmethod(wrapper)

class _Py2PNML (object) :
    "Convert Python objects to PNML"
    def __init__ (self) :
        self._handle = {"bool" : self._direct,
                        "chr" : self._direct,
                        "complex" : self._direct,
                        "dict" : self._dict,
                        "file" : self._pickle,
                        "float" : self._direct,
                        "int" : self._direct,
                        "list" : self._collection,
                        "long" : self._direct,
                        "set" : self._collection,
                        "str" : self._direct,
                        "tuple" : self._collection,
                        "unichr" : self._direct,
                        "unicode" : self._direct}
    def __call__ (self, doc, obj) :
        if hasattr(obj, "__pnmldump__") :
            return obj.__pnmldump__(doc)
        else :
            handle = self._handle.get(type(obj).__name__, self._pickle)
            return handle(doc, obj)
    def _direct (self, doc, obj) :
        result = doc.createElement("object")
        result.setAttribute("method", "direct")
        result.setAttribute("class", type(obj).__name__)
        result.setAttribute("value", repr(obj))
        return result
    def _pickle (self, doc, obj) :
        result = doc.createElement("object")
        result.setAttribute("method", "pickle")
        result.setAttribute("value", repr(pickle.dumps(obj)))
        return result
    def _collection (self, doc, obj) :
        result = doc.createElement("object")
        result.setAttribute("method", "collection")
        result.setAttribute("class", type(obj).__name__)
        for value in obj :
            result.appendChild(self(doc, value))
        return result
    def _dict (self, doc, obj) :
        result = doc.createElement("object")
        result.setAttribute("method", "dict")
        result.setAttribute("class", type(obj).__name__)
        for key, value in obj.items() :
            item = doc.createElement("item")
            k = doc.createElement("key")
            k.appendChild(self(doc, key))
            v = doc.createElement("value")
            v.appendChild(self(doc, value))
            item.appendChild(k)
            item.appendChild(v)
            result.appendChild(item)
        return result

_obj_to_pnml = _Py2PNML()
def obj_to_pnml (doc, obj) :
    """Convert an object to a DOM tree

    @param doc: the DOM document to which the tree will be added
    @type doc: C{xml.dom.minidom.Document}
    @param obj: the object to be serialized
    @type obj: any type
    """
    return _obj_to_pnml(doc, obj)

def load (doc) :
    """Load objects from a DOM document

    @param doc: the DOM document to parse
    @type doc: C{xml.dom.minidom.Document}
    @return: C{None} if no valid object has been found, an object if
      only one has been found, a tuple of objects otherwise
    @rtype: C{None} or C{object} or C{tuple}
    """
    if doc.documentElement.nodeName != "pnml" :
        raise ValueError, "not in PNML format"
    result = []
    for child in doc.documentElement.childNodes :
        result.append(pnml_to_obj(child))
    if len(result) == 0 :
        return None
    elif len(result) == 1 :
        return result[0]
    else :
        return tuple(result)

def _clean (doc) :
    "Remove empty text nodes from C{doc}"
    queue = [doc]
    while len(queue) > 0 :
        node = queue.pop(0)
        if node.hasChildNodes() :
            for child in node.childNodes[:] :
                try :
                    if child.data.strip() == "" :
                        node.removeChild(child).unlink()
                    else :
                        queue.append(child)
                except :
                    queue.append(child)

def loads (string) :
    """Load objects from a PNML string

    @param string: the PNML string to parse
    @type string: C{str}
    @return: C{None} if no valid object has been found, an object if
      only one has been found, a tuple of objects otherwise
    @rtype: C{None} or C{object} or C{tuple}
    """
    doc = xml.dom.minidom.parseString(string.strip())
    _clean(doc)
    return load(doc)

class _PNML2Py (object) :
    "Convert PNML to Python objects"
    def __init__ (self) :
        self._modules = {"" : __import__("__builtin__")}
        self.__doc__ = _PNML2Py.__doc__
    def __call__ (self, node) :
        if node.hasAttribute("plugins") :
            return self[node].__pnmlload__(node)
        elif node.nodeName == "object" :
            return getattr(self, "_" + string(node.getAttribute("method")))(node)
        else :
            raise ValueError, "tag '%s' not handled" % node.nodeName
    def __getitem__ (self, node) :
        path = string(node.getAttribute("plugins"))
        name = string(node.getAttribute("class"))
        if path not in self._modules :
            if ":" not in path :
                self._modules[path] = __import__(path, None, None, ["__name__"])
            else :
                base, plugins = path.split(":", 1)
                module = __import__(base, None, None, ["__name__"])
                for plug in plugins.split(":") :
                    module = snakes.plugins.load(plug, module)
                self._modules[path] = module
        return getattr(self._modules[path], name)
    def _direct (self, node) :
        cls = self[node]
        return cls(eval(string(node.getAttribute("value"))))
    def _pickle (self, node) :
        return pickle.loads(eval(string(node.getAttribute("value"))))
    def _collection (self, node) :
        result = []
        cls = self[node]
        for child in node.childNodes :
            col.append(self(child))
        return cls(result)
    def _dict (self, node) :
        result = {}
        cls = self[node]
        for item in node.childNodes :
            key = pnml_to_obj(item.getElementsByTagName("key")[0].firstChild)
            value = pnml_to_obj(item.getElementsByTagName("value")[0].firstChild)
            result[key] = value
        return cls(result)

_pnml_to_obj = _PNML2Py()
def pnml_to_obj (node) :
    """Build an object from a DOM node

    @param node: a DOM node
    @type node: C{xml.dom.minidom.Node}
    @return: the corresponding object
    @rtype: C{object}
    """
    return _pnml_to_obj(node)
