"""Typing system for places (and transitions guards).

Types are contraint checkers to verify whether a value is in the type
or not. For instance:

>>> 5 in tInteger
True
>>> 4.3 in tInteger
False

The typechecking of several values is possible using the type as a
function. It fails if one value is not in the type.

>>> tInteger(5, 4, 6)
True
>>> tInteger(5, 4, 6.0)
False

Types can be composed in order to build more complexe types. For
example:

>>> 8 in (tInteger & ~Range(1, 5))
True
>>> 3 in (tInteger & ~Range(1, 5))
False

The various compositions are the same as sets operations, plus the
complement: C{&}, C{|}, C{-}, C{^} and C{~} (complement).

Various types are predefined (like C{tInteger}) the others can be
constructed using the various classes in the module. Prefefined types
are:

 - C{tAll}: any value is in the type
 - C{tNothing}: empty type
 - C{tString}: string values
 - C{tList}: lists values
 - C{tInteger}: integer values
 - C{tNatural}: non-negative integers
 - C{tPositive}: strictly positive integers
 - C{tFloat}: float values
 - C{tNumber}: integers or float values
 - C{tDict}: Python's C{dict} values
 - C{tNone}: the single value C{None}
 - C{tBoolean}: C{True} or C{False}
 - C{tTuple}: tuple values
 - C{tPair}: tuples of length two
 - C{tBlackToken}: the single value C{"dot"}

Types with finitely many elements can be iterated:

>>> list(sorted(iter(CrossProduct(Range(0, 10, 2), tBoolean) & ~OneOf((4, True), (6, False)))))
[(0, False), (0, True), (2, False), (2, True), (4, False), (6, True), (8, False), (8, True)]
>>> list(iter(OneOf(1, 2, 3)))
[1, 2, 3]
>>> iter(tInteger)
Traceback (most recent call last):
...
TypeError: ...
>>> iter(~OneOf(1, 2, 3))
Traceback (most recent call last):
...
TypeError: ...
"""

from snakes.data import cross

def _iterable (obj, *types) :
    for t in types :
        try :
            iter(t)
        except :
            def __iterable__ () :
                raise ValueError, "iteration over non-sequence"
            obj.__iterable__ = __iterable__
            break

class Type (object) :
    """Base class for all types.

    Implement operations C{&}, C{|}, C{-}, C{^} and C{~} to build new
    types. Also implement the typechecking of several values. All the
    subclasses should implement the method C{__contains__} to
    typecheck a single object.
    """
    #C# skip
    def __init__ (self) :
        raise NotImplementedError, "abstract class"
    def __and__ (self, other) :
        """Intersection type.

        @param other: the other type in the intersection
        @type other: C{Type}
        @return: the intersection of both types
        @rtype: C{Type}
        """
        return _And(self, other)
    def __or__ (self, other) :
        """Union type.

        @param other: the other type in the union
        @type other: C{Type}
        @return: the union of both types
        @rtype: C{Type}
        """
        return _Or(self, other)
    def __sub__ (self, other) :
        """Substraction type.

        @param other: the other type in the substraction
        @type other: C{Type}
        @return: the type C{self} minus the type C{other}
        @rtype: C{Type}
        """
        return _Sub(self, other)
    def __xor__ (self, other) :
        """Disjoint union type.

        @param other: the other type in the disjoint union
        @type other: C{Type}
        @return: the disjoint union of both types
        @rtype: C{Type}
        """
        return _Xor(self, other)
    def __invert__ (self) :
        """Complementary type.

        @return: the complementary type
        @rtype: C{Type}
        """
        return _Invert(self)
    #C# skip
    def __iterable__ (self) :
        pass
    def __call__ (self, *values) :
        """Typecheck values.

        @param values: values that have to be checked
        @type values: C{object}
        @return: C{True} if all the values are in the types, C{False}
          otherwise
        @rtype: C{bool}
        """
        for v in values :
            if v not in self :
                return False
        return True

class _And (Type) :
    "Interstection of two types"
    def __init__ (self, left, right) :
        self._left = left
        self._right = right
        _iterable(self, left)
    def __repr__ (self) :
        return "(%s & %s)" % (repr(self._left), repr(self._right))
    def __contains__ (self, value) :
        """Check wether a value is in the type.

        @param value: the value to check
        @type value: C{object}
        @return: C{True} if C{value} is in the type, C{False} otherwise
        @rtype: C{bool}
        """
        return (value in self._left) and (value in self._right)
    def __iter__ (self) :
        self.__iterable__()
        for value in self._left :
            if value in self._right :
                yield value

class _Or (Type) :
    "Union of two types"
    def __init__ (self, left, right) :
        self._left = left
        self._right = right
        _iterable(self, left, right)
    def __repr__ (self) :
        return "(%s | %s)" % (repr(self._left), repr(self._right))
    def __contains__ (self, value) :
        """Check wether a value is in the type.

        @param value: the value to check
        @type value: C{object}
        @return: C{True} if C{value} is in the type, C{False} otherwise
        @rtype: C{bool}
        """
        return (value in self._left) or (value in self._right)
    def __iter__ (self) :
        self.__iterable__()
        for value in (self._left & self._right) :
            yield value
        for value in (self._left ^ self._right) :
            yield value

class _Sub (Type) :
    "Subtyping by difference"
    def __init__ (self, left, right)  :
        self._left = self._right
        _iterable(self, left)
    def __repr__ (self) :
        return "(%s - %s)" % (repr(self._left), repr(self._right))
    def __contains__ (self, value) :
        """Check wether a value is in the type.

        @param value: the value to check
        @type value: C{object}
        @return: C{True} if C{value} is in the type, C{False} otherwise
        @rtype: C{bool}
        """
        return (value in self._left) and (value not in self._right)
    def __iter__ (self) :
        self.__iterable__()
        for value in self._left :
            if value not in self._right :
                yield value

class _Xor (Type) :
    "Exclusive union of two types"
    def __init__ (self, left, right) :
        self._left = left
        self._right = right
        _iterable(self, left, right)
    def __repr__ (self) :
        return "(%s ^ %s)" % (repr(self._left), repr(self._right))
    def __contains__ (self, value) :
        """Check wether a value is in the type.

        @param value: the value to check
        @type value: C{object}
        @return: C{True} if C{value} is in the type, C{False} otherwise
        @rtype: C{bool}
        """
        if value in self._left :
            return value not in self._right
        else :
            return value in self._right
    def __iter__ (self) :
        self.__iterable__()
        for value in self._left :
            if value not in self._right :
                yield value
        for value in self._right :
            if value not in self._left :
                yield value

class _Invert (Type) :
    "Complement of a type"
    def __init__ (self, base) :
        self._base = base
        _iterable(self, None)
    def __repr__ (self) :
        return "(~%s)" % repr(self._base)
    def __contains__ (self, value) :
        """Check wether a value is in the type.

        @param value: the value to check
        @type value: C{object}
        @return: C{True} if C{value} is in the type, C{False} otherwise
        @rtype: C{bool}
        """
        return (value not in self._base)

class _All (Type) :
    "A type allowing for any value"
    def __init__ (self) :
        pass
    def __and__ (self, other) :
        return other
    def __or__ (self, other) :
        return self
    def __sub__ (self, other) :
        return ~other
    def __xor__ (self, other) :
        return ~other
    def __invert__ (self) :
        return tNothing
    def __repr__ (self) :
        return "tAll"
    def __contains__ (self, value) :
        """Check wether a value is in the type.

        @param value: the value to check
        @type value: C{object}
        @return: C{True}
        @rtype: C{bool}
        """
        return True
    def __call__ (self, *values) :
        """Typecheck values.

        @param values: values that have to be checked
        @type values: any objet
        @return: C{True}
        """
        return True

class _Nothing (Type) :
    "A types with no value"
    def __init__ (self) :
        pass
    def __and__ (self, other) :
        return self
    def __or__ (self, other) :
        return other
    def __sub__ (self, other) :
        return self
    def __xor__ (self, other) :
        return other
    def __invert__ (self) :
        return tAll
    def __repr__ (self) :
        return "tNothing"
    def __contains__ (self, value) :
        """Check wether a value is in the type.

        @param value: the value to check
        @type value: C{object}
        @return: C{False}
        @rtype: C{bool}
        """
        return False
    def __call__ (self, *values) :
        """Typecheck values.

        @param values: values that have to be checked
        @type values: any objet
        @return: C{False}
        """
        return False
    def __iter__ (self) :
        pass

class Instance (Type) :
    """A type whose values are all instances of one class.

    >>> [1, 2] in Instance(list)
    True
    >>> (1, 2) in Instance(list)
    False
    """
    def __init__ (self, _class) :
        """
        @param _class: the class of instance
        @type _class: any class
        """
        self._class = _class
    def __contains__ (self, value) :
        """Check wether a value is in the type.

        @param value: the value to check
        @type value: C{object}
        @return: C{True} if C{value} is in the type, C{False} otherwise
        @rtype: C{bool}
        """
        return isinstance(value, self._class)
    def __repr__ (self) :
        """
        >>> repr(Instance(str))
        'Instance(str)'
        """
        return "Instance(%s)" % self._class.__name__

class TypeCheck (Type) :
    """A type whose values are accepted by a given function.

    >>> def odd (val) :
    ...     return type(val) is int and (val % 2) == 1
    >>> 3 in TypeCheck(odd)
    True
    >>> 4 in TypeCheck(odd)
    False
    >>> 3.0 in TypeCheck(odd)
    False
    """
    def __init__ (self, checker, iterate=None) :
        """
        @param checker: a function that checks one value and returns
          C{True} if it is in te type and C{False} otherwise
        @type checker: C{function(value)->bool}
        @param iterate: C{None} or an iterator over the values of the
          type
        @type iterate: C{None} or C{iterator}
        """
        self._check = checker
        self._iterate = iterate
    def __iter__ (self) :
        """
        >>> def odd (val) :
        ...     return type(val) is int and (val % 2) == 1
        >>> i = iter(TypeCheck(odd))
        Traceback (most recent call last):
          ...
        ValueError: type not iterable
        >>> def odd_iter () :
        ...     i = 1
        ...     while True :
        ...         yield i
        ...         yield -i
        ...         i += 2
        >>> i = iter(TypeCheck(odd, odd_iter))
        >>> i.next(), i.next(), i.next()
        (1, -1, 3)
        """
        try :
            return iter(self._iterate())
        except TypeError :
            raise ValueError, "type not iterable"
    def __contains__ (self, value) :
        """Check wether a value is in the type.

        @param value: the value to check
        @type value: C{object}
        @return: C{True} if C{value} is in the type, C{False} otherwise
        @rtype: C{bool}
        """
        return self._check(value)
    def __repr__ (self) :
        """
        >>> def odd (val) :
        ...     return type(val) is int and (val % 2) == 1
        >>> repr(TypeCheck(odd))
        'TypeCheck(odd)'
        >>> def odd_iter () :
        ...     i = 1
        ...     while True :
        ...         yield i
        ...         yield -i
        ...         i += 2
        >>> repr(TypeCheck(odd, odd_iter))
        'TypeCheck(odd, odd_iter)'
        """
        if self._iterate is None :
            return "%s(%s)" % (self.__class__.__name__,
                               self._check.func_name)
        else :
            return "%s(%s, %s)" % (self.__class__.__name__,
                                   self._check.func_name,
                                   self._iterate.func_name)

class OneOf (Type) :
    """A type whose values are explicitely enumerated.

    >>> 3 in OneOf(1, 2, 3, 4, 5)
    True
    >>> 0 in OneOf(1, 2, 3, 4, 5)
    False
    """
    def __init__ (self, *values) :
        """
        @param values: the enumeration of the values in the type
        @type values: any objects
        """
        self._values = values
    def __contains__ (self, value) :
        """Check wether a value is in the type.

        @param value: the value to check
        @type value: C{object}
        @return: C{True} if C{value} is in the type, C{False} otherwise
        @rtype: C{bool}
        """
        return value in self._values
    def __repr__ (self) :
        """
        >>> repr(OneOf(1, 2, 3, 4, 5))
        'OneOf(1, 2, 3, 4, 5)'
        """
        return "OneOf(%s)" % ", ".join([repr(val) for val in self._values])
    def __iter__ (self) :
        """
        >>> list(iter(OneOf(1, 2, 3, 4, 5)))
        [1, 2, 3, 4, 5]
        """
        return iter(self._values)

class Collection (Type) :
    """A type whose values are a given container, holding items of a
    given type and ranging in a given interval.

    >>> [0, 1.1, 2, 3.3, 4] in Collection(Instance(list), tNumber, 3, 10)
    True
    >>> [0, 1.1] in Collection(Instance(list), tNumber, 3, 10) #too short
    False
    >>> [0, '1.1', 2, 3.3, 4] in Collection(Instance(list), tNumber, 3, 10) #wrong item
    False
    """
    def __init__ (self, collection, items, min=None, max=None) :
        """
        @param collection: the collection type
        @type collection: any container type
        @param items: the type of the items
        @type items: any type
        @param min: the smallest allowed value
        @type min: any value in C{items}
        @param max: the greatest allowed value
        @type max: any value in C{items}
        """
        self._collection = collection
        self._items = items
        self._max = max
        self._min = min
    def __contains__ (self, value) :
        """Check wether a value is in the type.

        @param value: the value to check
        @type value: C{object}
        @return: C{True} if C{value} is in the type, C{False} otherwise
        @rtype: C{bool}
        """
        if value not in self._collection :
            return False
        try :
            len(value)
            iter(value)
        except TypeError :
            return False
        if (self._min is not None) and (len(value) < self._min) :
            return False
        if (self._max is not None) and (len(value) > self._max) :
            return False
        for item in value :
            if item not in self._items :
                return False
        return True
    def __repr__ (self) :
        """
        >>> repr(Collection(Instance(list), tNumber, 3, 10))
        'Collection(Instance(list), (Instance(int) | Instance(float)), min=3, max=10)'
        """
        if (self._min is None) and (self._max is None) :
            return "Collection(%s, %s)" % (repr(self._collection),
                                           repr(self._items))
        elif self._min is None :
            return "Collection(%s, %s, max=%s)" % (repr(self._collection),
                                                   repr(self._items),
                                                   repr(self._max))
        elif self._max is None :
            return "Collection(%s, %s, min=%s)" % (repr(self._collection),
                                                   repr(self._items),
                                                   repr(self._min))
        else :
            return "Collection(%s, %s, min=%s, max=%s)" % (repr(self._collection),
                                                           repr(self._items),
                                                           repr(self._min),
                                                           repr(self._max))

def List (items, min=None, max=None) :
    """Shorthand for instantiating C{Collection}

    >>> List(tNumber, min=3, max=10)
    Collection(Instance(list), (Instance(int) | Instance(float)), min=3, max=10)
    """
    return Collection(Instance(list), items, min, max)

def Tuple (items, min=None, max=None) :
    """Shorthand for instantiating C{Collection}

    >>> Tuple(tNumber, min=3, max=10)
    Collection(Instance(tuple), (Instance(int) | Instance(float)), min=3, max=10)
    """
    return Collection(Instance(tuple), items, min, max)

def Set (items, min=None, max=None) :
    """Shorthand for instantiating C{Collection}

    >>> Set(tNumber, min=3, max=10)
    Collection(Instance(set), (Instance(int) | Instance(float)), min=3, max=10)
    """
    return Collection(Instance(set), items, min, max)

class Mapping (Type) :
    """A type whose values are mapping (eg, C{dict})

    >>> {'Yes': True, 'No': False} in Mapping(tString, tAll)
    True
    >>> {True: 1, False: 0} in Mapping(tString, tAll)
    False
    """
    def __init__ (self, keys, items, _dict=dict) :
        """
        @param keys: the type for the keys
        @type keys: any type
        @param items: the type for the items
        @type items: any type
        @param _dict: the class that mapping must be instances of
        @type _dict: any C{dict} like class
        """
        self._keys = keys
        self._items = items
        self._dict = _dict
    def __contains__ (self, value) :
        """Check wether a value is in the type.

        @param value: the value to check
        @type value: C{object}
        @return: C{True} if C{value} is in the type, C{False} otherwise
        @rtype: C{bool}
        """
        if not isinstance(value, self._dict) :
            return False
        for key, item in value.items() :
            if key not in self._keys :
                return False
            if item not in self._items :
                return True
        return True
    def __repr__ (self) :
        """
        >>> repr(Mapping(tString, tAll))
        'Mapping(Instance(str), tAll)'
        """
        if self._dict is dict :
            return "Mapping(%s, %s)" % (repr(self._keys),
                                        repr(self._items))
        else :
            return "Mapping(%s, %s, %s)" % (repr(self._keys),
                                            repr(self._items),
                                            self._dict.__name__)

class Range (Type) :
    """A type whose values are in a given range

    Notice that ranges are not built into the memory so that huge
    values can be used.

    >>> 3 in Range(1, 2**128, 2)
    True
    >>> 4 in Range(1, 2**128, 2)
    False
    >>> 4 in range(1, 2**128, 2) # Python's builtin
    Traceback (most recent call last):
    ...
    OverflowError: range() result has too many items
    """
    def __init__ (self, first, last, step=1) :
        """The values are those that the builtin C{range(first, last, step)}
        would return."""
        self._first, self._last, self._step = first, last, step
    def __contains__ (self, value) :
        """Check wether a value is in the type.

        @param value: the value to check
        @type value: C{object}
        @return: C{True} if C{value} is in the type, C{False} otherwise
        @rtype: C{bool}
        """
        return ((self._first <= value <= self._last)
                and ((value - self._first) % self._step == 0))
    def __repr__ (self) :
        """
        >>> repr(Range(1, 2**128, 2))
        'Range(1, 340282366920938463463374607431768211456, 2)'
        """
        if self._step == 1 :
            return "Range(%s, %s)" % (self._first, self._last)
        else :
            return "Range(%s, %s, %s)" % (self._first, self._last, self._step)
    def __iter__ (self) :
        """
        >>> list(iter(Range(1, 10, 3)))
        [1, 4, 7]
        """
        return iter(xrange(self._first, self._last, self._step))

class Greater (Type) :
    """A type whose values are greater than a minimum.

    The minimum and the checked values can be of any type as soon as
    they can be compared with C{>}.

    >>> 6 in Greater(3)
    True
    >>> 3 in Greater(3)
    False
    >>> 4 in Greater(None) # Python yields C{True} on: C{4 > None}
    True
    """
    def __init__ (self, min) :
        """
        @param min: the minimal allowed value
        @type min: any type for which Python will be able to apply C{>}
        """
        self._min = min
    def __contains__ (self, value) :
        """Check wether a value is in the type.

        @param value: the value to check
        @type value: C{object}
        @return: C{True} if C{value} is in the type, C{False} otherwise
        @rtype: C{bool}
        """
        try :
            return value > self._min
        except :
            return False
    def __repr__ (self) :
        """
        >>> repr(Greater(3))
        'Greater(3)'
        """
        return "Greater(%s)" % repr(self._min)

class GreaterOrEqual (Type) :
    """A type whose values are greater or equal than a minimum.

    See the description of C{Greater}
    """
    def __init__ (self, min) :
        self._min = min
    def __contains__ (self, value) :
        """Check wether a value is in the type.

        @param value: the value to check
        @type value: C{object}
        @return: C{True} if C{value} is in the type, C{False} otherwise
        @rtype: C{bool}
        """
        try :
            return value >= self._min
        except :
            False
    def __repr__ (self) :
        """
        >>> repr(GreaterOrEqual(3))
        'GreaterOrEqual(3)'
        """
        return "GreaterOrEqual(%s)" % repr(self._min)

class Less (Type) :
    """A type whose values are less than a maximum.

    See the description of C{Greater}
    """
    def __init__ (self, max) :
        self._max = max
    def __contains__ (self, value) :
        """Check wether a value is in the type.

        @param value: the value to check
        @type value: C{object}
        @return: C{True} if C{value} is in the type, C{False} otherwise
        @rtype: C{bool}
        """
        return value < self._max
    def __repr__ (self) :
        """
        >>> repr(Less(3))
        'Less(3)'
        """
        return "Less(%s)" % repr(self._max)

class LessOrEqual (Type) :
    """A type whose values are less than or equal to a maximum.

    See the description of C{Greater}
    """
    def __init__ (self, max) :
        self._max = max
    def __contains__ (self, value) :
        """Check wether a value is in the type.

        @param value: the value to check
        @type value: C{object}
        @return: C{True} if C{value} is in the type, C{False} otherwise
        @rtype: C{bool}
        """
        return value <= self._max
    def __repr__ (self) :
        """
        >>> repr(LessOrEqual(3))
        'LessOrEqual(3)'
        """
        return "LessOrEqual(%s)" % repr(self._max)

class CrossProduct (Type) :
    """A type whose values are tuples, each component of them being in
    given types. The resulting type is the cartesian cross product of
    the compound types.

    >>> (1, 3, 5) in CrossProduct(Range(1, 10), Range(1, 10, 2), Range(1, 10))
    True
    >>> (2, 4, 6) in CrossProduct(Range(1, 10), Range(1, 10, 2), Range(1, 10))
    False
    """
    def __init__ (self, *types) :
        """
        @param types: the types of each component of the allowed tuples
        @type types: C{Type}
        """
        self._types = types
        _iterable(self, *types)
    def __repr__ (self) :
        """
        >>> repr(CrossProduct(Range(1, 10), Range(1, 10, 2), Range(1, 10)))
        'CrossProduct(Range(1, 10), Range(1, 10, 2), Range(1, 10))'
        """
        return "CrossProduct(%s)" % ", ".join([repr(t) for t in self._types])
    def __contains__ (self, value) :
        """Check wether a value is in the type.

        @param value: the value to check
        @type value: C{object}
        @return: C{True} if C{value} is in the type, C{False} otherwise
        @rtype: C{bool}
        """
        if (not isinstance(value, tuple)) and (len(self._types) == len(value)) :
            return False
        for item, t in zip(value, self._types) :
            if not item in t :
                return False
        return True
    def __iter__ (self) :
        """A cross product is iterable if so are all its components.

        >>> list(iter(CrossProduct(Range(1, 3), Range(3, 5))))
        [(1, 3), (1, 4), (2, 3), (2, 4)]
        >>> iter(CrossProduct(Range(1, 100), tAll))
        Traceback (most recent call last):
        ...
        ValueError: iteration over non-sequence
        """
        self.__iterable__()
        return cross(self._types)

tAll        = _All()
tNothing    = _Nothing()
tString     = Instance(str)
tList       = List(tAll)
tInteger    = Instance(int)
tNatural    = tInteger & GreaterOrEqual(0)
tPositive   = tInteger & Greater(0)
tFloat      = Instance(float)
tNumber     = tInteger|tFloat
tDict       = Instance(dict)
tNone       = OneOf(None)
tBoolean    = OneOf(True, False)
tTuple      = Tuple(tAll)
tPair       = Tuple(tAll, min=2, max=2)
tBlackToken = OneOf("dot")
